import pytest

from tackle.models import Context
from tackle.render import render_string
from tackle import tackle
from tackle.utils.files import read_config_file


RENDERABLES = [
    # This test relates to https://github.com/pallets/jinja/issues/1580 where if we have
    # a variable that collides with a jinja globals it is ignored
    # https://github.com/robcxyz/tackle-box/issues/19
    ({'dict': {'stuff': 'things'}}, '{{dict}}', {'stuff': 'things'}),
    ({'namespace': {'stuff': 'things'}}, '{{namespace}}', {'stuff': 'things'}),
    # Normal
    ({'adict': {'stuff': 'things'}}, '{{adict}}', {'stuff': 'things'}),
    ({'list': ['stuff', 'things']}, '{{list}}', ['stuff', 'things']),
    # Dashes don't work ->
    # ({'a-dash': 'dash'}, '{{a-dash}}', 'dash'),
]


@pytest.mark.parametrize("context,raw,expected_output", RENDERABLES)
def test_render_string_inputs(context, raw, expected_output):
    context = Context(public_context=context)
    output = render_string(context=context, raw=raw)

    assert output == expected_output


FILES = [
    # TODO: Not asserting - RM
    # ('multi-line-block.yaml', 'multi-line-block-output.yaml'),
    ('call-function.yaml', 'call-function-output.yaml'),
]


@pytest.mark.parametrize("file,expected_output", FILES)
def test_render_files(change_curdir_fixtures, file, expected_output):
    o = tackle(file)
    expected_output = read_config_file(expected_output)
    assert o == expected_output


def test_render_hook_call_multiple(change_curdir_fixtures):
    """
    Check that when we run a hook multiple times, that we don't carry over the prior
    hook calls arguments which may not be instantiated on the second hook call. This
    happens when the hook that was called is left in the jinja env's globals and not
    removed which makes the hook not an `unknown_variable` so it uses the prior args.
    """
    o = tackle('multiple-hook-renders.yaml')
    assert o['first'] == "foo,stuff"
    assert o['second'] == "2.txt"



from typing import Union

from tackle.models import BaseHook, Context, Field
from tackle.parser import walk_sync


class BlockHook(BaseHook):
    """
    Hook for blocks of hooks. This is a special case where `items` are parsed like a
     normal context with the added benefit of maintaining a `temporary_context` so that
     items on the same level can be accessed in memory / rendered. Normally executed
     via a macro with an arrow. This the only hook the core parser is aware of as it is
     parsing.
    """

    hook_type: str = 'block'
    items: Union[dict, list] = Field(
        ..., description="Items to be parsed like a normal input."
    )

    skip_output: bool = True
    _render_exclude = {'items'}

    def exec(self) -> Union[dict, list]:
        if self.temporary_context is None:
            self.temporary_context = {}

        tmp_context = Context(
            public_hooks=self.public_hooks,
            private_hooks=self.private_hooks,
            public_context=self.public_context,
            private_context=self.private_context,
            temporary_context=self.temporary_context,
            existing_context=self.existing_context,
            input_context=self.items,
            key_path=self.key_path.copy(),
            key_path_block=self.key_path.copy(),
            no_input=self.no_input,
            calling_directory=self.calling_directory,
            calling_file=self.calling_file,
            verbose=self.verbose,
        )
        walk_sync(context=tmp_context, element=self.items.copy())

        return self.public_context



"""Main `tackle` CLI."""
import sys
import argparse

from tackle import __version__
from tackle.utils.log import configure_logger
from tackle.main import tackle
from tackle.utils.command import unpack_args_kwargs_list


def main(raw_args=None):
    """Run a tackle box."""
    if raw_args is None:
        raw_args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description="tackle-box is a DSL for creating declarative CLIs. Call tackle "
        "against files, directories, or repos with yaml/toml/json tackle "
        "files."
    )

    parser.add_argument(
        dest='input_string',
        nargs="?",
        type=str,
        default=None,
        help="The input to parse as a source either to a local file / directory or "
        "remote location - ie github.com/robcxyz/tackle-demos or "
        "robcxyz/tackle-demos. \n"
        # "If no source is matched then the tool defaults to searching in parent "
        # "directories for a .tackle.yaml file using this argument in for running "
        # "that file.",
    )
    parser.add_argument(
        '--no-input',
        '-n',
        action='store_true',
        help="Do not prompt for anything resulting in default choices.",
    )
    parser.add_argument(
        '--verbose', '-v', action='store_true', help="Execute in verbose mode."
    )
    parser.add_argument(
        '--print', '-p', action='store_true', help="Print the resulting output."
    )
    parser.add_argument(
        '--print-format',
        '-pf',
        default=None,
        dest='print_format',
        type=str,
        metavar="",
        help="The format to print to output. Defaults to json.",
    )
    parser.add_argument(
        '--checkout',
        '-c',
        dest='checkout',
        type=str,
        metavar="",
        help="The branch or version to checkout if a remote source is given.",
    )
    parser.add_argument(
        '--latest',
        '-l',
        action='store_true',
        help="When using version controlled providers (ie in github), use the latest "
        "commit in the default branch.",
    )
    parser.add_argument(
        '--directory',
        '-d',
        default=None,
        type=str,
        metavar="",
        help="The directory to run from if given a remote source.",
    )
    parser.add_argument(
        '--file',
        '-f',
        type=str,
        metavar="",
        dest="context_file",
        help="The file to run. Only relevent for remote sources as otherwise you can "
        "just give full path to file as input.",
    )
    parser.add_argument(
        '--find-in-parent',
        '-fp',
        action='store_true',
        help="Search for target in parent directory. Only relevant for local targets.",
    )
    parser.add_argument(
        '--override',
        '-o',
        type=str,
        default=None,
        metavar="",
        help="A string to a file to use as overrides when parsing a tackle file or "
        "some dict to use with keys to additionally use when parsing a file.",
    )
    parser.add_argument(
        '--version', action='version', version=f'tackle-box {__version__}'
    )

    args, unknown_args = parser.parse_known_args(raw_args)

    # Handle printing variable which isn't needed in core logic
    print_format = args.print_format
    if print_format is not None:
        print_enabled = True
        if print_format not in ['json', 'yaml', 'toml']:
            print(
                f"Invalid `--print-format` / `-pf` input \"{print_format}\". "
                f"Must be one of json (default) / yaml / toml."
            )
            sys.exit(1)
    else:
        print_enabled = args.print
        print_format = 'json'
    del args.print_format
    del args.print

    # Breakup the unknown arguments which are added later into the runtime's args/kwargs
    global_args, global_kwargs, global_flags = unpack_args_kwargs_list(unknown_args)

    configure_logger(
        stream_level='DEBUG' if getattr(args, "verbose", False) else 'INFO',
        debug_file=getattr(args, "debug_file", None),
    )

    output = tackle(
        **vars(args),
        global_args=global_args,
        global_kwargs=global_kwargs,
        global_flags=global_flags,
    )
    if print_enabled:
        if isinstance(output, (dict, list)):
            if print_format == 'json':  # noqa
                import json

                print(json.dumps(dict(output)))
            elif print_format == 'yaml':
                from ruamel.yaml import YAML

                yaml = YAML()
                yaml.dump(output, sys.stdout)
            elif print_format == 'toml':
                from toml import dumps as toml_dumps

                print(toml_dumps(dict(output)))
        else:
            print(output)


if __name__ == "__main__":
    main(sys.argv[1:])



import pathlib
import os
import re
import subprocess
from collections.abc import MutableMapping

from InquirerPy import prompt

from pydantic import BaseModel, ValidationError, validator
from typing import Optional

from tackle.utils.paths import work_in
from tackle.models import BaseHook, Field


class Repo(BaseModel):
    """Repo class for parsing dict inputs for repo."""

    src: str
    branch: Optional[str] = None
    tag: Optional[str] = None


# https://stackoverflow.com/a/6027615/12642712
def flatten_repo_tree(d, parent_key=''):
    """Flatten a dict to so that keys become nodes in a path."""
    items = []
    for k, v in d.items():
        new_key = os.path.join(parent_key, k)

        try:
            repo = Repo(**v)
        except (ValidationError, TypeError):
            repo = None

        if repo:
            items.append((new_key, v))
        elif isinstance(v, MutableMapping):
            items.extend(flatten_repo_tree(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


class MetaGitHook(BaseHook):
    """
    Hook to create meta repo, ie a repo that has many other repos within it. See
     [meta repo definition.](https://notes.burke.libbey.me/metarepo/#:~:text=a%20%E2%80%9CMetarepo%E2%80%9D.-,Metarepo%20Architecture,branch%20commit%20for%20each%20repository.)
     Describe the repo structure with a map and when this hook is called, it allows the
     users to operate git commands on that repo structure.
    """

    __slots__ = ('first_run',)
    # Per https://github.com/samuelcolvin/pydantic/issues/655 for private vars

    hook_type: str = 'meta_repo'
    command: str = Field(None, description="The git command to run.")
    repo_tree: dict = Field(
        None,
        description="A map of repos to clone with the keys indicating the path "
        "and values for the repo targets. See examples.",
    )
    protocol: str = Field(
        "https",
        description="Either https or ssh. Useful when using abbreviated repo "
        "references.",
    )
    token: str = Field(
        None, description="A token to use with https abbreviated repo references."
    )
    base_url: str = Field(
        "github.com",
        description="The base url to use with https abbreviated repo " "references.",
    )
    git_org: str = Field(
        None,
        description="The git org to override with https abbreviated repo "
        "references.",
    )

    _doc_tags: list = ["experimental"]
    _issue_numbers: list = []

    @validator('token')
    def no_token_with_ssh_protocol(cls, v, values, **kwargs):
        if v and values['protocol'] == 'ssh':
            raise ValueError("Can't supply token with ssh protocol.")
        else:
            return v

    @validator('base_url')
    def update_base_url_if_token_exists(cls, v, values, **kwargs):
        if values['token']:
            return values['token'] + ':x-oauth-basic@' + v
        else:
            return v

    def get_repo_prefix(self, repo):
        """Return a string to prefix url."""
        if repo.startswith("https") or repo.startswith("git"):
            # Case where it is provided
            return ""
        if self.protocol == "https":
            return f"https://{self.base_url}/"
        if self.protocol == "ssh":
            return f"git@{self.base_url}:"

    def get_git_command(self, branch, repo, folder_name):
        """Build string to run git command."""
        if self.command == 'clone':
            if not branch:
                return f"git {self.command} {self.get_repo_prefix(repo)}{repo} {folder_name}"
            else:
                return (
                    f"git {self.command} {self.get_repo_prefix(repo)}{repo} "
                    f"{folder_name} -b {branch}"
                )
        elif self.command == 'pull':
            return f"git {self.command}"

        # if not branch:
        #     return (
        #         f"git {self.command} {self.get_repo_prefix(repo)}{repo} {folder_name}"
        #     )
        # else:
        #     return (
        #         f"git {self.command} {self.get_repo_prefix(repo)}{repo} "
        #         f"{folder_name} -b {branch}"
        #     )

    def execute_git_command(self, branch, repo, folder_name):
        """Execute the git command."""
        cmd = self.get_git_command(branch, repo, folder_name)
        p = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        output, err = p.communicate()

        if err:
            print(f"{err} for repo='{repo}' in folder='{folder_name}'.")

    def try_https_then_ssh(self, branch, repo, folder_name):
        """When running first git command, tries with https then falls back to ssh."""
        try:
            self.execute_git_command(branch, repo, folder_name)
            object.__setattr__(self, 'first_run', False)
        except Exception:
            try:
                self.protocol = "ssh"
                self.execute_git_command(branch, repo, folder_name)
            except Exception as e:
                print(f"{e} for repo='{repo}' - Neither https or ssh works. Skipping.")

    def run_git_command(self, repo, folder_name, branch=None):
        """Run the git command."""
        if self.first_run:
            self.try_https_then_ssh(branch, repo, folder_name)
        self.execute_git_command(branch, repo, folder_name)

    def update_repo_name(self, v):
        """Update the repo name."""
        if re.compile(r"(https:\/\/[\w\.]+)|(git@[\w\.]+)").match(v):
            return v

        git_parts = v.split('/')
        if len(git_parts) == 1:
            if not self.git_org:
                print(
                    f"No git org declared - only one item declared in {v}, "
                    f"must be in form org/repo for abbreviated declarations. Skipping."
                )
                return
            else:
                return f"{self.git_org}/{git_parts[0]}"
        if len(git_parts) == 2:
            return f"{git_parts[0]}/{git_parts[1]}"
        else:
            from tackle.utils.dicts import get_key_from_key_path

            key = get_key_from_key_path(self.key_path)
            print(f"Malformed repo name '{v}' in '{key}' key. Skipping.")

    def prompt_repo_choices(self):
        """Prompt the user to select which items to operate on."""
        choices = [
            {"name": f"{k} --> {v}", "checked": True} for k, v in self.repo_tree.items()
        ]
        question = {
            'type': 'checkbox',
            'name': 'tmp',
            'message': f"Which repos do you want to {self.command.split(' ')[0]}?",
            'choices': choices,
        }
        repo_choices = prompt([question])['tmp']
        self.repo_tree = {
            k: v for k, v in self.repo_tree.items() if f"{k} --> {v}" in repo_choices
        }

    def get_command(self):
        """Get the git command to run."""
        if not self.no_input:
            git_commands = [
                '<custom input>',
                'clone',
                'pull',
                'branch',
            ]

            question = {
                'type': 'list',
                'name': 'tmp',
                'message': "What to do?",
                'choices': git_commands,
            }
            self.command = prompt([question])['tmp']

            # Extra input
            if self.command == '<custom input>':
                message = "Enter git command: "
                question = {
                    'type': 'input',
                    'name': 'tmp',
                    'message': message,
                }
                self.command = prompt([question])['tmp']
            if self.command == 'branch':
                message = "Enter git branch: "
                question = {
                    'type': 'input',
                    'name': 'tmp',
                    'message': message,
                }
                self.command = "branch " + prompt([question])['tmp']

        else:
            # print(
            #     f"No command given in mete_hook for key = {self.key}.  "
            #     f"Defaults to clone."
            # )
            self.command = 'clone'

    def exec(self):
        """Run the hook."""
        object.__setattr__(self, 'first_run', True)

        if not self.command:
            self.get_command()

        # Flatten the repo tree = `{k1: {k2: v}}` -> `[{k1/k2: v}]` where `k1/k2`
        # is the path to operate (ie clone, etc) on the repo object `v` that can be
        # a string or dict
        if self.repo_tree:
            self.repo_tree = flatten_repo_tree(self.repo_tree)

        # Select is a bool that can override whether to prompt
        if not self.no_input:
            self.prompt_repo_choices()

        # Parse flattened repo tree from above
        for k, v in self.repo_tree.items():
            # Create dir to base path if it does not exist
            path, folder_name = os.path.split(k)
            if not os.path.isdir(path):
                pathlib.Path(path).mkdir(parents=True, exist_ok=True)

            # Handle string inputs
            if isinstance(v, str):
                v = self.update_repo_name(v)
                if path != "":
                    with work_in(path):
                        self.run_git_command(v, folder_name)
                else:
                    self.run_git_command(v, folder_name)

            if isinstance(v, dict):
                r = Repo(**v)
                r.src = self.update_repo_name(r.src)
                if path != "":
                    with work_in(path):
                        self.run_git_command(r.src, folder_name, r.branch)
                else:
                    self.run_git_command(r.src, folder_name, r.branch)



from typing import Union, Optional

from tackle import BaseHook, Field
from tackle.utils.dicts import nested_get, encode_key_path


class DistinctHook(BaseHook):
    """Hook for getting distinct items from a list."""

    hook_type: str = 'distinct'
    # fmt: off
    src: Union[list, str] = Field(
        ..., description="A list to get the distinct values from or a str with "
                         "separators or list for a key path to the element to rewrite "
                         "the distinct items back to the context.")
    src_is_key_path: bool = Field(
        False, description="If the src is a list and is meant to be a key path.")
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # fmt: on

    args: list = ['src']

    def exec(self) -> Optional[list]:
        if isinstance(self.src, str) or self.src_is_key_path:
            self.src = encode_key_path(self.src, self.sep)
            # When appending within a block, we try to append to output dict then
            # fallback on trying to append to the existing context
            try:
                self.src = nested_get(
                    element=self.public_context,
                    keys=self.src,
                )
            except KeyError:
                self.src = nested_get(
                    element=self.existing_context,
                    keys=self.src,
                )
        # TODO: Raise better error if retrieved key path is not a list
        distinct_list = list(set(self.src))
        return distinct_list



"""Tests dict input objects for `tackle.providers.system.hooks.copy` module."""
import os
import shutil
from tackle.main import tackle
import pytest


@pytest.fixture()
def clean_files(change_dir):
    """Clean the run."""
    yield
    files = ['output', os.path.join('output', 'foo')]
    for f in files:
        if os.path.isfile(f):
            os.remove(f)
        if os.path.isdir(f):
            shutil.rmtree(f)


def test_provider_system_hook_file(change_dir, clean_files):
    o = tackle('tackle.yaml')
    assert os.path.isdir('output')
    assert os.path.exists(o['join'])


def test_provider_system_hook_file_args(change_dir, clean_files):
    o = tackle('tackle.yaml')
    assert os.path.isdir('output')
    assert os.path.exists(o['join'])



from typing import Union, Any, Optional

from tackle import BaseHook, Field
from tackle.utils.dicts import nested_get, encode_key_path, get_target_and_key


class ListAppendHook(BaseHook):
    """Hook for updating dict objects with items."""

    hook_type: str = 'append'
    # fmt: off
    src: Union[list, str] = Field(
        ..., description="A list append to and output the result or a str with "
                         "separators or list for a key path to the element to append "
                         "to within the context.")
    src_is_key_path: bool = Field(
        False, description="If the src is a list and is meant to be a key path.")
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # item: Union[str, list, dict, int, float, bool] = Field(
    #     ..., description="An item to append to a list."
    # )
    item: Any = Field(
        ..., description="An item to append to a list."
    )
    # fmt: on

    args: list = ['src', 'item']

    def exec(self) -> Optional[list]:
        if isinstance(self.src, str) or self.src_is_key_path:
            key_path = encode_key_path(self.src, self.sep)
            # When appending within a block, we try to append to output dict then
            # fallback on trying to append to the existing context
            target_context, set_key_path = get_target_and_key(self, key_path=key_path)

            try:
                self.src = nested_get(
                    element=target_context,
                    keys=set_key_path,
                )
            except KeyError:
                self.src = nested_get(
                    element=self.temporary_context,
                    keys=set_key_path,
                )

            self.src.append(self.item)
        else:
            self.src.append(self.item)
            return self.src



import sys
from typing import Any
from tackle import BaseHook, Field

try:
    from rich import print
    from rich.pretty import pprint
except ImportError:
    from pprint import pprint


class PrintHook(BaseHook):
    """
    Hook for printing an input and returning the output.
    [Link](https://docs.python.org/3/library/functions.html#print)
    """

    hook_type: str = 'print'

    objects: Any = Field(None, description="The objects to print.")
    sep: str = Field(' ', description="Separator between printed objects.")
    end: str = Field('\n', description="What to print at the end")
    flush: bool = Field(False, description="No clue.")

    args: list = ['objects']

    def exec(self) -> None:
        print(self.objects, sep=self.sep, end=self.end, flush=self.flush)


class PprintHook(BaseHook):
    """
    Wraps python pprint builtin.
    [Link](https://docs.python.org/3/library/pprint.html#pprint.PrettyPrinter)
    """

    hook_type: str = 'pprint'

    # fmt: off
    objects: Any = Field(None, description="The object to be pretty printed.")
    indent: int = Field(
        1,
        description="Specifies the amount of indentation added for each nesting level.")
    width: int = Field(
        80,
        description="Specifies the desired maximum number of characters per line in "
                    "the output. If a structure cannot be formatted within the width "
                    "constraint, a best effort will be made.")
    depth: int = Field(
        None,
        description="Controls the number of nesting levels which may be printed; if "
                    "the data structure being printed is too deep, the next contained "
                    "level is replaced by .... By default, there is no constraint on "
                    "the depth of the objects being formatted.")
    compact: bool = Field(
        False,
        description="Impacts the way that long sequences (lists, tuples, sets, etc) "
                    "are formatted. If compact is false (the default) then each item "
                    "of a sequence will be formatted on a separate line. If compact "
                    "is true, as many items as will fit within the width will be "
                    "formatted on each output line.")
    sort_dicts: bool = Field(
        True,
        description="If sort_dicts is true (the default), dictionaries will be "
                    "formatted with their keys sorted, otherwise they will display in "
                    "insertion order.")
    underscore_numbers: bool = Field(
        False,
        description="If underscore_numbers is true, integers will be formatted with "
                    "the _ character for a thousands separator, otherwise underscores "
                    "are not displayed (the default).")
    # fmt: on

    args: list = ['objects']

    def exec(self) -> None:
        try:
            if sys.version_info.minor < 8:
                if 'rich' in sys.modules:
                    pprint(
                        self.objects,
                        max_length=self.width,
                    )
                else:
                    pprint(
                        self.objects,
                        indent=self.indent,
                        width=self.width,
                        compact=self.compact,
                    )
            elif sys.version_info.minor < 10:
                if 'rich' in sys.modules:
                    # TODO: Line up these docs
                    # https://rich.readthedocs.io/en/stable/reference/pretty.html?highlight=pprint#rich.pretty.pprint
                    # https://github.com/robcxyz/tackle-box/issues/57
                    pprint(
                        self.objects,
                        max_length=self.width,
                    )
                else:
                    pprint(
                        self.objects,
                        indent=self.indent,
                        width=self.width,
                        compact=self.compact,
                        sort_dicts=self.sort_dicts,
                    )
            elif sys.version_info.minor >= 10:
                if 'rich' in sys.modules:
                    pprint(
                        self.objects,
                        max_length=self.width,
                    )
                else:
                    pprint(
                        self.objects,
                        indent=self.indent,
                        width=self.width,
                        compact=self.compact,
                        sort_dicts=self.sort_dicts,
                        underscore_numbers=self.underscore_numbers,
                    )
        except TypeError as e:
            # TODO: Raise better exception?
            raise e
        return



from typing import Any

from pydantic import Field

from tackle import BaseHook
from tackle.utils.dicts import encode_key_path, nested_set, get_target_and_key


class SetKeyHook(BaseHook):
    """
    Hook for setting a key based on a key path which is a list with keys and numbers
     for indexes in a list.
    """

    hook_type: str = 'set'
    # fmt: off
    path: Any = Field(..., description="A list or string with a separator for the path to the value you want to update with strings for keys and ints for indexes in the list.")
    value: Any = Field(..., description="The value to update the key with.")
    sep: str = Field('/', description="For string paths, a separator for key path.")
    args: list = ['path', 'value']
    # fmt: on

    def exec(self):
        """Run the hook."""
        target_context, set_key_path = get_target_and_key(
            self, key_path=encode_key_path(self.path, self.sep)
        )

        nested_set(
            element=target_context,
            keys=set_key_path,
            value=self.value,
        )



from typing import Any

from tackle.models import BaseHook, Field


class LiteralHook(BaseHook):
    """Literally return the input."""

    hook_type: str = 'literal'
    input: Any = Field(..., description="Any variable input.")

    args: list = ['input']

    def exec(self) -> Any:
        return self.input



import pytest
import os

from tackle import tackle
from tackle.utils.files import read_config_file
from tackle.exceptions import EmptyBlockException

FIXTURES = [
    ('single-level.yaml', 'single-level-output.yaml'),
    ('nested-dict.yaml', 'nested-dict-output.yaml'),
    ('nested-for.yaml', 'nested-for-output.yaml'),
    # Consider RM - Too much stuff
    # ('ansible-parse-call.yaml', 'ansible-parse-output.yaml')
]


@pytest.mark.parametrize("input,output", FIXTURES)
def test_parser_blocks_validate_output(chdir, input, output):
    """Test blocks."""
    chdir(os.path.join("fixtures", "blocks"))

    tackle_output = tackle(input)
    assert tackle_output == read_config_file(output)


ERROR_SOURCES = [
    ("empty-block-exception.yaml", EmptyBlockException),
]


@pytest.mark.parametrize("input_file,exception", ERROR_SOURCES)
def test_parser_raises_exceptions(chdir, input_file, exception):
    """Validate that exceptions are raised."""
    chdir(os.path.join("fixtures", "blocks"))
    with pytest.raises(exception):
        tackle(input_file)


def test_parser_calling_directory_preserve(change_curdir_fixtures):
    """Validate that the calling_directory param is carried over from hook calls."""
    output = tackle('calling-context.yaml')
    assert output['call']['call']['calling_file'] == 'calling-context.yaml'


def test_parser_list_to_block_macro(change_curdir_fixtures):
    os.chdir('macros')
    output = tackle('list-block.yaml')
    assert isinstance(output['foo'][1], list)


def test_parser_compact_hook_call_macro(change_curdir_fixtures):
    """Check that embedded compact hooks are called appropriately."""
    os.chdir('macros')
    output = tackle('compact-hook-macro.yaml')
    assert output['compact'] == 'things'


# def test_parser_compact_hook_call_macro(change_curdir_fixtures):
#     os.chdir('macros')
#     output = tackle('compact-hook-macro2.yaml')
#     assert output['compact'] == 'things'



"""High level tests for parser logic."""
import pytest

from tackle.main import tackle
from tackle.utils.files import read_config_file

FIXTURES = [
    ('map.yaml', 'map-output.yaml'),
    ('map-lists.yaml', 'map-lists-output.yaml'),
    ('arg-types.yaml', 'arg-types-output.yaml'),
    ('empty-elements.yaml', 'empty-elements.yaml'),
    ('map-root.yaml', 'map-root-output.yaml'),
    ('private-hooks.yaml', 'private-hooks-output.yaml'),
    ('outer-tackle.yaml', 'outer-tackle-output.yaml'),
    ('outer-tackle-list.yaml', 'outer-tackle-list-output.yaml'),
    ('merge-simple.yaml', 'merge-simple-output.yaml'),
    ('merge-petstore-compact.yaml', 'petstore.yaml'),
    ('private-context.yaml', 'private-context-output.yaml'),
    # Non tackle things
    ('k8s-deployment.yaml', 'k8s-deployment.yaml'),
    ('ansible-playbook.yaml', 'ansible-playbook.yaml'),
    ('docker-compose.yml', 'docker-compose.yml'),
    ('list-list.yaml', 'list-list.yaml'),
    ('var-hook.yaml', 'var-hook-output.yaml'),
    ('toml.toml', 'toml.yaml'),
    # # Broken
    # # TODO: https://github.com/robcxyz/tackle-box/issues/52
    # ('bug-mixed-flags.yaml', 'bug-mixed-flags.yaml'),
    # ('document-hooks.yaml', 'document-hooks-expected.yaml'),
]


@pytest.mark.parametrize("fixture,expected_output", FIXTURES)
def test_main_expected_output(change_curdir_fixtures, fixture, expected_output):
    """Input equals output."""
    expected_output = read_config_file(expected_output)
    output = tackle(fixture)
    assert output == expected_output


def test_parser_ruamel_braces(change_curdir_fixtures):
    """
    Validate super hack for ruamel parsing error where `stuff->: {{things}}`
    (no quotes), ruamel interprets as:
    'stuff': ordereddict([(ordereddict([('things', None)]), None)]).
    """
    output = tackle('ruamel-parsing-error-braces.yaml', verbose=True)
    assert output['stuff'] == 'things'
    assert output['a']['b'] is None
    assert output['one'] == 'two'



"""Utilities mainly used in helping `modes` like replay and others."""
import json
from ruamel.yaml import YAML
from ruamel.yaml.composer import ComposerError
import os
import logging

from tackle.exceptions import (
    ContextDecodingException,
    UnsupportedBaseFileTypeException,
    TackleImportError,
)
from tackle.utils.paths import make_sure_path_exists

logger = logging.getLogger(__name__)


def get_file_name(replay_dir, template_name, suffix='yaml'):
    """Get the name of file."""
    suffix = '.' + suffix if not template_name.endswith('.' + suffix) else ''
    file_name = '{}{}'.format(template_name, suffix)
    return os.path.join(replay_dir, file_name)


def dump(output_dir, output_name, output_dict, dump_output='yaml'):
    """Write json data to file."""
    if not make_sure_path_exists(output_dir):
        raise IOError('Unable to create replay dir at {}'.format(output_dir))

    replay_file = get_file_name(output_dir, output_name, dump_output)

    if dump_output == 'json':
        with open(replay_file, 'w') as f:
            json.dump(output_dict, f, indent=2)
    if dump_output in ['yaml', 'yml']:
        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        with open(replay_file, 'w') as f:
            yaml.dump(output_dict, f)


def read_config_file(file, file_extension=None):
    """Read files into objects."""
    if not file_extension:
        file_extension = file.split('.')[-1]

    if not os.path.exists(file):
        raise FileNotFoundError(f"Can't find the file {file}.")

    logger.debug(
        'Using \"{}\" as input file and \"{}\" as file extension'.format(
            file, file_extension
        )
    )
    try:
        if file_extension == 'json':
            with open(file) as f:
                config = json.load(f)
            return config
        elif file_extension in ('yaml', 'yml'):
            # Try normal and then for documents that will output as list
            yaml = YAML()
            try:
                with open(file, encoding='utf-8') as f:
                    config = yaml.load(f)
                return config
            except ComposerError:
                output = []
                with open(file, encoding='utf-8') as f:
                    for doc in yaml.load_all(f.read()):
                        output.append(doc)
                return output
        elif file_extension == 'toml':
            try:
                import toml
            except ImportError:
                raise TackleImportError(
                    f"Error parsing {file} No toml package installed. Install it with "
                    "`pip install toml` and try again."
                ) from None
            with open(file) as f:
                data = toml.load(f)
            return data

        else:
            raise UnsupportedBaseFileTypeException(
                'Unable to parse file {}. Error: Unsupported extension (json/yaml only)'
                ''.format(file)
            )  # noqa

    except ValueError as e:
        # JSON decoding error.  Let's throw a new exception that is more
        # friendly for the developer or user.
        message = (
            f'JSON decoding error while loading "{file}".  Decoding'
            f' error details: "{str(e)}"'
        )
        raise ContextDecodingException(message) from None


def apply_overwrites_to_inputs(input, overwrite_dict):
    """Modify the given context in place based on the overwrite_context."""
    for variable, overwrite in overwrite_dict.items():
        if variable not in input:
            # Do not include variables which are not used in the template
            continue

        context_value = input[variable]

        if isinstance(context_value, list):
            # We are dealing with a choice variable
            if overwrite in context_value:
                # This overwrite is actually valid for the given context
                # Let's set it as default (by definition first item in list)
                context_value.remove(overwrite)
                context_value.insert(0, overwrite)
        else:
            # Simply overwrite the value for this variable
            input[variable] = overwrite



"""Collection of tests around tackle box's declarative command-line interface."""

import pytest
from tackle.cli import main

COMMANDS = [
    'tackle --help',
    # 'help --foo bar',
    # 'docs',
]


@pytest.fixture(params=COMMANDS)
def basic_commands(request):
    """Pytest fixture return both version invocation options."""
    return request.param


@pytest.mark.parametrize("command", COMMANDS)
def test_cli_commands(change_curdir_fixtures, command, capsys):
    """Asser output comes out of cli."""
    main(['tackle-hello.yaml'])
    assert 'Hello world!' in capsys.readouterr().out



"""All exceptions are defined here."""
import inspect
import os
import sys
from typing import TYPE_CHECKING

from tackle.utils.dicts import get_readable_key_path

if TYPE_CHECKING:
    from typing import Type, Union
    from pydantic.main import ModelMetaclass
    from tackle.models import Context, BaseFunction, BaseHook, BaseContext


def raise_unknown_hook(context: 'Context', hook_type: str, method: bool = None):
    """Raise an exception for when there is a missing hook with some context."""
    if method:
        type_ = ["method", "hook"]
        extra = ""  # TODO
        # extra = "Available methods include"
    else:
        type_ = ["hook", "providers"]
        if context.verbose:
            available_hooks = "".join(
                sorted([str(i) for i in context.provider_hooks.keys()])
            )
            extra = f'Available hooks = {available_hooks}'

        else:
            extra = "Run the application with `--verbose` to see available hook types."

    raise UnknownHookTypeException(
        f"The {type_[0]}=\"{hook_type}\" is not available in the {type_[1]}. " + extra,
        context=context,
    )


class ContributionNeededException(Exception):
    """
    Special exception to point users (particularly of the Windows variety) to contribute
     a fix with context that includes a link to the file needing upgrade.
    """

    def __init__(self, extra_message=None):
        # self.path_to_code = link_to_code
        self.extra_message = extra_message
        self.message = (
            f"Unimplemented / needs to be built "
            f"- PLEASE HELP -> "
            f"https://github.com/robcxyz/tackle-box/blob/main/{self.file_location()}"
            f"\n{self.extra_message}"
        )
        super().__init__(self.message)

    def __new__(cls, *args, **kwargs):
        new_instance = super(ContributionNeededException, cls).__new__(
            cls, *args, **kwargs
        )
        stack_trace = inspect.stack()
        created_at = '%s:%d' % (stack_trace[1][1], stack_trace[1][2])
        new_instance.created_at = created_at
        return new_instance

    def file_location(self):
        path_list = os.path.normpath(self.created_at).split(os.sep)
        for i, v in enumerate(path_list):
            if v == 'tackle':
                break
        github_path_list = path_list[i:-1] + [
            path_list[-1].split(':')[0] + "#L" + path_list[-1].split(':')[1]
        ]
        return os.path.join(*github_path_list)


#
# Base exceptions - Ones that are subclassed later
#


class TackleHookCallException(Exception):
    """Base hook call exception class. Subclassed within providers."""

    def __init__(self, extra_message: str, hook: 'Type[BaseHook]' = None):
        self.message = (
            f"Error parsing input_file='{hook.calling_file}' at "
            f"key_path='{get_readable_key_path(key_path=hook.key_path)}' \n"
            f"{extra_message}"
        )
        if not hook.verbose:
            sys.tracebacklimit = 0
        super().__init__(self.message)


class HookUnknownChdirException(TackleHookCallException):
    """
    Exception for a hook call with a chdir method.

    Raised when chdir is to an unknown directory.
    """


#
# Function call exceptions
#
class TackleFunctionCallException(Exception):
    """Base hook call exception class."""

    def __init__(
        self,
        extra_message: str,
        function: 'Union[BaseFunction, Type[ModelMetaclass]]',
    ):
        self.message = (
            f"Error parsing input_file='{function.calling_file}' at "
            f"key_path='{get_readable_key_path(key_path=function.key_path)}' \n"
            f"{extra_message}"
        )
        if not function.verbose:
            sys.tracebacklimit = 0
        super().__init__(self.message)


class FunctionCallException(TackleFunctionCallException):
    """
    Exception for when walking a function input context.

    Raised in function calls.
    """


#
# Parser exceptions
#
class TackleParserException(Exception):
    """Base parser exception class."""

    def __init__(self, extra_message: str, context: 'Union[Context, BaseContext]'):
        self.message = (
            f"Error parsing input_file='{context.current_file}' at "
            f"key_path='{get_readable_key_path(key_path=context.key_path)}' \n"
            f"{extra_message}"
        )
        if not context.verbose:
            sys.tracebacklimit = 0
        super().__init__(self.message)


class HookCallException(TackleParserException):
    """
    Exception for an unknown field in a hook.

    Raised when field has been provided not declared in the hook type.
    """


class HookParseException(TackleParserException):
    """
    Exception for an unknown field in a hook.

    Raised when field has been provided not declared in the hook type.
    """


class AppendMergeException(TackleParserException):
    """
    Excpetion when user tries to merge from for loop (ie hook --for [] --merge).

    Raised when merging the output of a hook call.
    """


class TopLevelMergeException(TackleParserException):
    """
    Excpetion when user tries to merge from top level (ie t->: literal foo --merge).

    Raised when merging the output of a hook call.
    """


class EmptyBlockException(TackleParserException):
    """
    Exception when a block (ie a compact hook without a hook type - key->: if:/for:)
    is declared without any contents.

    Raised when handling empty blocks.
    """


class UnknownArgumentException(TackleParserException):
    """
    Exception unknown argument when supplied via a hook call.

    Raised when a hook is called with an unknown argument in it's mapping (ie missing
    `_args` param in the hook definition.
    """


class UnknownTemplateVariableException(TackleParserException):
    """
    Exception for an unknown templatable argument.

    Raised when rendering variables.
    """


class MissingTemplateArgsException(TackleParserException):
    """
    Exception with missing arguments in hook call within template.

    Raised when rendering variables.
    """


class TooManyTemplateArgsException(TackleParserException):
    """
    Exception with missing arguments in hook call within template.

    Raised when rendering variables.
    """


class UnknownHookTypeException(TackleParserException):
    """
    Exception for unknown hook type.

    Raised when a hook type is not available from the providers.
    """


class MalformedTemplateVariableException(TackleParserException):
    """
    Exception for a malformed templatable argument.

    Raised when rendering variables.
    """


#
# Parser input exceptions
#
class TackleParserInputException(Exception):
    """Base input parser exception class."""

    def __init__(self, extra_message: str, context: 'Context' = None):
        self.message = (
            f"Error parsing input_file='{context.input_file}' \n" f"{extra_message}"
        )
        if not context.verbose:
            sys.tracebacklimit = 0
        super().__init__(self.message)


# class TackleFileInitialParsingException(TackleParserInputException):
#     """
#     Exception when base tackle file is empty.
#
#     Raised when calling files / providers.
#     """


class EmptyTackleFileException(TackleParserInputException):
    """
    Exception when base tackle file is empty.

    Raised when calling files / providers.
    """


class UnknownSourceException(TackleParserInputException):
    """
    Exception for ambiguous source.

    Raised when tackle cannot determine which directory is the project
    template, e.g. more than one dir appears to be a template dir.
    """


class UnknownInputArgumentException(TackleParserInputException):
    """
    Exception for unknown extra arguments.

    Raised when tackle cannot determine what the extra argument means.
    """


#
# Util exceptions -> No context
#
class TackleException(Exception):
    """
    Base exception class.

    All Tackle-specific exceptions should subclass this class.
    """

    def __init__(self, message: str):
        self.message = message
        sys.tracebacklimit = 0
        super().__init__(self.message)


class InvalidConfiguration(TackleException):
    """
    Exception for invalid configuration file.

    Raised if the global configuration file is not valid YAML or is
    badly constructed.
    """


# class UnknownRepoType(TackleException):
#     """
#     Exception for unknown repo types.
#
#     Raised if a repo's type cannot be determined.
#     """


class VCSNotInstalled(TackleException):
    """
    Exception when version control is unavailable.

    Raised if the version control system (git or hg) is not installed.
    """


class VersionNotFoundError(TackleException):
    """
    Exception when a version to a provider is given but does not exist.

    Raised if a version is specified but does not exist on the remote.
    """


class UnsupportedBaseFileTypeException(TackleException):
    """
    Exception for when a none json / yaml file are read

    Raised if the base file that is being called is not json / yaml.
    """


class ContextDecodingException(TackleException):
    """
    Exception for failed JSON decoding.

    Raised when a project's JSON context file can not be decoded.
    """


class RepositoryNotFound(TackleException):
    """
    Exception for missing repo.

    Raised when the specified tackle repository doesn't exist.
    """


class ContextFileNotFound(TackleException):
    """
    Exception for missing context file.

    Raised when the specified context file doesn't exist.
    """


class RepositoryCloneFailed(TackleException):
    """
    Exception for un-cloneable repo.

    Raised when a tackle template can't be cloned.
    """


class InvalidZipRepository(TackleException):
    """
    Exception for bad zip repo.

    Raised when the specified tackle repository isn't a valid
    Zip archive.
    """


#
# Function create exceptions
#
class TackleFunctionCreateException(Exception):
    """Base hook call exception class."""

    def __init__(
        self, extra_message: str, function_name: str, context: 'Context' = None
    ):
        self.message = (
            f"Error creating hook='{function_name}' in file="
            f"'{context.calling_file}', {extra_message}"
        )
        if not context.verbose:
            sys.tracebacklimit = 0
        super().__init__(self.message)


class EmptyFunctionException(TackleFunctionCreateException):
    """
    Exception when a function is declared without any input.

    Happens when a tackle file is parsed.
    """


class MalformedFunctionFieldException(TackleFunctionCreateException):
    """
    Exception when functions with field inputs of type dict are not formatted
    appropriately.

    Happens when a tackle file is parsed.
    """


class ShadowedFunctionFieldException(TackleFunctionCreateException):
    """
    Exception when functions with field inputs of type dict are not formatted
    appropriately.

    Happens when a tackle file is parsed.
    """


class TackleGeneralException(Exception):
    """Base hook call exception class."""

    def __init__(self, message: str):
        sys.tracebacklimit = 0
        super().__init__(message)


class NoInputOrParentTackleException(TackleGeneralException):
    """
    Exception when no input has been given, nor is there a parent tackle.

    Raised when a hook is called with an unknown argument in it's mapping (ie missing
    `_args` param in the hook definition.
    """


class TackleImportError(TackleGeneralException):
    """
    Exception when functions with field inputs of type dict are not formatted
    appropriately.

    Happens when a tackle file is parsed.
    """


class TackleFileInitialParsingException(TackleGeneralException):
    """
    Exception when parsing a tackle file.

    Raised when first parsing a file.
    """



import datetime
from datetime import timezone

from tackle import BaseHook, Field


class DateTimeNowHook(BaseHook):
    """Hook for updating dict objects with items."""

    hook_type: str = 'date_now'
    format: str = Field(
        "%m-%d-%Y",
        description="Date time formatting per [the official docs](https://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior)",
    )

    timestamp: bool = Field(
        False, description="Output as timestamp. Precedence over format."
    )
    utc: bool = Field(False, description="Output in UTC.")

    args: list = ['format']

    def exec(self) -> str:

        if self.timestamp:
            if self.utc:
                dt = datetime.datetime.now(timezone.utc)
                utc_time = dt.replace(tzinfo=timezone.utc)
                now = utc_time.timestamp()
            else:
                now = datetime.datetime.now().timestamp()
        else:
            if self.utc:
                now = datetime.datetime.utcnow().strftime(f'{self.format}')
            else:
                now = datetime.datetime.now().strftime(f'{self.format}')
        return now



import pytest
from tackle import tackle
from tackle.exceptions import HookCallException


def test_hook_match_loop(change_dir):
    """Run a loop of different match cases."""
    output = tackle('loop.yaml')
    # Assertions in fixture
    assert output


# TODO: https://github.com/robcxyz/tackle-box/issues/66
#  Add merge to loop functions
# def test_hook_match_block_loop(change_dir):
#     """Check that we can merge from a looped block into a looped match."""
#     output = tackle('block-loop.yaml')
#     # Assertions in fixture
#     assert output

# TODO: https://github.com/robcxyz/tackle-box/issues/66
#  Fix basic block macro functionality
# def test_hook_match_block_if(change_dir):
#     """Check that any kind of first level block makes sense."""
#     output = tackle('block-if.yaml')
#     # TODO: Assertions in fixture
#     assert output


def test_hook_match_cases(change_dir):
    """Run the source and check that the hooks imported the demo module."""
    output = tackle('cases.yaml')
    assert output['matched_dict'] == 'this'
    assert output['fallback_dict'] == 'foo'


def test_hook_match_value_list(change_dir):
    """
    Edge case where in match hooks one can have a single value trying to be merged into
    a temporary context which does not make sense.
    """
    output = tackle('value-list.yaml')
    # Assertions in file
    assert output


def test_hook_match_value_lists(change_dir):
    """
    Edge case where in match hooks one can have a single value trying to be merged into
    a temporary context which does not make sense.
    """
    o = tackle('lists.yaml')
    assert o


def test_hook_match_value_wrong_hook_type(change_dir):
    """
    Edge case where in match hooks one can have a single value trying to be merged into
    a temporary context which does not make sense.
    """
    with pytest.raises(HookCallException):
        tackle('wrong-hook-type.yaml')



"""Main package for tackle box."""
__version__ = "0.4.1"  # x-release-please-version

from tackle.models import BaseHook
from tackle.models import Field
from tackle.main import tackle


__all__ = [
    'tackle',
]



import os
import random
from pathlib import Path
import shutil
from distutils.dir_util import copy_tree
from typing import List, Union, Any, Optional

from tackle.models import BaseHook, Field
from tackle.exceptions import HookCallException


def create_directory_tree(src, dst):
    """Create the file path if it does not exist."""
    if isinstance(src, list) and not os.path.isdir(dst):
        Path(dst).mkdir(parents=True, exist_ok=True)
    elif isinstance(src, str) and not os.path.isdir(os.path.dirname(dst)):
        Path(os.path.basename(dst)).mkdir(parents=True, exist_ok=True)
    elif isinstance(src, dict):
        raise NotImplementedError


def expand_path(path):
    """Expand the path for user."""
    if isinstance(path, str):
        return os.path.abspath(os.path.expandvars(os.path.expanduser(path)))
    if isinstance(path, list):
        return [
            os.path.abspath(os.path.expandvars(os.path.expanduser(f))) for f in path
        ]


class CopyHook(BaseHook):
    """Hook coying a file/files or directory/directories to a location."""

    hook_type: str = 'copy'
    src: Union[List, str] = Field(
        ..., description="String or list of sources, either a directories or files"
    )
    dst: str = Field(..., description="The destination to copy to.")
    create_path: bool = True

    args: list = ['src', 'dst']

    def exec(self) -> None:
        self.dst = os.path.abspath(self.dst)

        if self.create_path:
            create_directory_tree(self.src, self.dst)

        if isinstance(self.src, str):
            self.src = os.path.abspath(os.path.expanduser(self.src))
            self._copy_file_or_dir(self.src, self.dst)
        elif isinstance(self.src, list):
            self.src = [os.path.abspath(os.path.expanduser(f)) for f in self.src]
            for i in self.src:
                self._copy_file_or_dir(i, os.path.join(self.dst, os.path.basename(i)))

        return None

    @staticmethod
    def _copy_file_or_dir(source, dest):
        if os.path.isdir(source):
            copy_tree(source, dest)
        elif os.path.isfile(source):
            shutil.copyfile(source, dest)


class MoveHook(BaseHook):
    """Hook  for moving a directory or directories to a location."""

    hook_type: str = 'move'
    src: Union[List, str] = Field(
        ..., description="String or list of sources, either directories or files"
    )
    dst: str = Field(..., description="String for path to copy to")
    create_path: bool = Field(
        True, description="Boolean to create the directory path if it does not exist."
    )

    args: list = ['src', 'dst']

    def exec(self) -> None:
        self.dst = os.path.abspath(self.dst)

        if self.create_path:
            create_directory_tree(self.src, self.dst)

        if isinstance(self.src, str):
            self.src = os.path.abspath(os.path.expanduser(self.src))
            shutil.move(self.src, self.dst)
        elif isinstance(self.src, list):
            self.src = [os.path.abspath(os.path.expanduser(f)) for f in self.src]
            for i in self.src:
                shutil.move(i, os.path.join(self.dst, os.path.basename(i)))

        return None


class RemoveHook(BaseHook):
    """Hook for removing a directory or directories."""

    hook_type: str = 'remove'
    path: Union[List, str] = Field(
        ..., description="String or list of paths to remove."
    )

    args: list = ['path']

    def remove_file_or_dir(self, path):
        if os.path.isfile(path):
            os.remove(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)
        else:
            raise HookCallException(f"Can't find path {path}.", context=self)

    def exec(self) -> None:
        if isinstance(self.path, str):
            self.path = os.path.abspath(os.path.expanduser(self.path))
            self.remove_file_or_dir(self.path)
        elif isinstance(self.path, list):
            self.path = [os.path.abspath(os.path.expanduser(f)) for f in self.path]
            for i in self.path:
                self.remove_file_or_dir(i)
        return None


# Source: https://codereview.stackexchange.com/questions/186130/file-shredding-secure-deletion-algorithm # noqa
def wipe(f, passes=30):
    """Overwrite a file with bytes."""
    if not os.path.isfile(f):
        return "Could not find the specified file!"
    with open(f, "ba+") as f2w:
        size = f2w.tell()
        for i in range(passes):
            f2w.seek(0)

            f2w.write(os.urandom(size))
    new_name = str(random.randint(1000, 1000000000000))
    os.rename(f, new_name)
    os.remove(new_name)
    return "Success!"


class ShredHook(BaseHook):
    """Hook for shredding file/files."""

    hook_type: str = 'shred'
    src: Union[List, str] = Field(
        ..., description="String or list of sources, either directories or files"
    )
    passes: int = Field(
        10, description="The number of passes to overwrite the location."
    )

    args: list = ['src', 'passes']

    def exec(self) -> None:
        if isinstance(self.src, str):
            self.src = os.path.abspath(os.path.expanduser(self.src))
            self.src = [self.src]
        for i in self.src:
            self.src = [os.path.abspath(os.path.expanduser(f)) for f in self.src]
            wipe(i, self.passes)


class ChmodHook(BaseHook):
    """Hook for changing the mode of a path or paths."""

    hook_type: str = 'chmod'
    path: Union[str, list] = Field(
        ..., description="String or list of paths, either directories or files."
    )
    mode: str = Field(
        ...,
        description="The mode to write to the files. See python's os package "
        "[chmod](https://docs.python.org/3/library/os.html#os.chmod) "
        "for more details.",
    )

    args: list = ['path', 'mode']

    def exec(self) -> None:
        if isinstance(self.path, str):
            self.path = [self.path]
        for i in self.path:
            mode = int(self.mode, 8)
            os.chmod(path=expand_path(i), mode=mode)

        return None


class CreateFileHook(BaseHook):
    """Hook to create an empty file - like touch."""

    hook_type: str = 'create_file'
    path: Union[str, list] = Field(
        ..., description="String or list of paths to create files at."
    )

    args: list = ['path']

    def exec(self) -> Union[str, list]:
        self.path = expand_path(self.path)
        if isinstance(self.path, str):
            self.path = [self.path]
        for i in self.path:
            Path(i).touch()

        return self.path


class FileHook(BaseHook):
    """Hook to read or write a file."""

    hook_type: str = 'file'
    path: str = Field(
        ..., description="Path to read a file or write file if contents are given."
    )
    contents: Any = Field(
        None, description="If writing to file, the contents to write."
    )

    args: list = ['path', 'contents']

    def exec(self) -> Optional[str]:
        self.path = expand_path(self.path)
        if self.contents is None:
            with open(self.path) as f:
                contents = f.read()
            return contents
        else:
            with open(self.path, 'w') as f:
                f.write(self.contents)
            return



from tackle.main import tackle


def test_provider_system_hook_block_tackle(change_dir):
    """Simple block test."""
    output = tackle('merge.yaml', no_input=True)

    assert output['stuff'] == 'here'


def test_provider_system_hook_block_embedded_blocks(change_dir):
    """Embedded with multiple blocks."""
    output = tackle('embedded-blocks.yaml', no_input=True)

    assert output['stuff'] == 'things'
    assert output['blocker']['things'] == 'things'


def test_provider_system_hook_block_embedded_blocks_2(change_dir):
    """Complex block."""
    output = tackle('embedded-lists.yaml')
    assert output['one'][1]['two'][0]['three']


def test_provider_system_hook_block_looped(change_dir):
    """With a for loop."""
    output = tackle('looped.yaml', no_input=True)

    assert len(output['blocker']) == 2
    assert output['blocker'][0]['compact'] == 'stuff'
    assert output['blocker'][1]['compact'] == 'things'
    assert 'priv_expanded' not in output
    assert 'priv_compact' not in output


def test_provider_system_hook_block_block_merge(change_dir):
    """Block with a merge."""
    output = tackle('block-merge.yaml', no_input=True)
    # TODO: Update tests with https://github.com/robcxyz/tackle-box/issues/51
    # assert output['things'] == 'here'
    assert output['stuff'] == 'things'


def test_provider_system_hook_block_block(change_dir):
    """Complex block."""
    output = tackle('block.yaml', no_input=True)
    # TODO: Update tests with https://github.com/robcxyz/tackle-box/issues/51
    assert output['block']['things'] == 'here'


def test_provider_system_hook_block_list(change_dir):
    """Macro re-written block, simple."""
    output = tackle('list.yaml')
    assert output['public'] == ['stuff', 'things']
    assert 'private' not in output


def test_provider_system_hook_block_looped_context(change_dir):
    """Check that a temp context is built."""
    output = tackle('looped-context.yaml')
    assert len(output['networks']) == 2
    assert output['networks'][0]['network_name'] == 'foo'


def test_provider_system_hook_block_list_block(change_dir):
    """Complex block."""
    output = tackle('block-list.yaml')
    assert 'private' not in output


def test_provider_system_hook_block_logic(change_dir):
    """Block with logic."""
    output = tackle('block-logic.yaml')
    assert output


def test_provider_system_hook_block_tmp_context(change_dir):
    """Check that a temp context is built."""
    output = tackle('tmp-context.yaml')
    assert output['public']



import os
from typing import Any

from pydantic import Field

from tackle import BaseHook
from tackle.utils.paths import find_in_parent


class FindInParentHook(BaseHook):
    """Hook to find the absolute path to a file or directory in parent directories."""

    hook_type: str = 'find_in_parent'
    target: str = Field(
        ..., description="The name of the file to find the absolute path to"
    )
    fallback: Any = Field(
        None, description="String to fallback on if the target is not found."
    )
    starting_dir: str = Field(
        '.',
        description="The starting directory to search from. Defaults to current working directory.",
    )

    args: list = ['target']

    def exec(self) -> str:
        return find_in_parent(
            dir=self.starting_dir, targets=[self.target], fallback=self.fallback
        )


class FindInChildHook(BaseHook):
    """Hook to find the absolute path to a file or directory in child directories."""

    hook_type: str = 'find_in_child'
    target: str = Field(
        ..., description="The name of the file to find the absolute path to"
    )
    fallback: Any = Field(
        None, description="String to fallback on if the target is not found."
    )
    starting_dir: str = Field(
        '.',
        description="The starting directory to search from. Defaults to current working directory.",
    )

    args: list = ['target']

    def exec(self) -> list:
        files = []
        for (dirpath, dirnames, filenames) in os.walk(self.starting_dir):
            files += [
                os.path.join(dirpath, file) for file in filenames if file == self.target
            ]

        if len(files) == 0:
            return self.fallback

        return files



from tackle.models import BaseHook, Field


class ExitHook(BaseHook):
    """Exit the parser with an exit code."""

    hook_type: str = 'exit'
    code: int = Field(0, description="The exit code.", render_by_default=True)

    args: list = ['code']

    def exec(self) -> None:
        exit(self.code)



# from .console import hooks
# from .paths import hooks

# import os
# import logging
# import subprocess
# import sys
# import importlib.machinery
# from pydantic import BaseModel, ConfigError
#
# from tackle.utils.paths import listdir_absolute
# from tackle.settings import settings
#
# from typing import List
#
# logger = logging.getLogger(__name__)
#
# file_path = os.path.dirname(os.path.abspath(__file__))
#
# native_providers = [
#     os.path.join(file_path, f)
#     for f in os.listdir(file_path)
#     if os.path.isdir(os.path.join(file_path, f)) and f != '__pycache__'
# ]
#
#
# class Provider(BaseModel):
#     """Base provider."""
#
#     path: str = None
#     hooks_path: str = None
#     hook_types: list = []
#     hook_modules: list = []
#
#     name: str = None  # defaults to os.path.basename(path)
#
#     src: str = None
#     version: str = None
#     requirements: list = []
#
#     def __init__(self, **data):
#         super().__init__(**data)
#
#
# class ProviderList(BaseModel):
#     """
#     Provider list object that on instantiation imports native (shipping wit tackle box),
#     settings (declared in ~/.tackle/settings.yaml), and local (in a local hooks
#     directory).
#     """
#
#     __root__: List[Provider] = []
#     # _functions: dict = None
#
#     @property
#     def hook_types(self):
#         hook_types = []
#         for i in self:
#             hook_types.append(i.name)
#         return hook_types
#
#     # @property
#     # def functions(self):
#     #     functions = {}
#     #     for i in self:
#     #         functions.append(i.functions)
#     #     return functions
#
#     def append(self, value: Provider):
#         """Override."""
#         self.__root__.append(value)
#
#     def __iter__(self):
#         """Override."""
#         return iter(self.__root__)
#
#     def __getitem__(self, item):
#         """Override."""
#         return self.__root__[item]
#
#     def __init__(self, **data):
#         super().__init__(**data)
#         """Import natve, settings, and local providers."""
#         # Import native providers in tackle.providers
#         self.import_paths(native_providers)
#         # Import any providers in settings
#         if settings.extra_providers:
#             self.import_paths(settings.extra_providers)
#         # Local `hooks` dir imported within parser
#
#     def import_paths(self, provider_paths):
#         """Iterate through paths and import them."""
#         for i in provider_paths:
#             self.append_from_path(i)
#
#     def append_from_path(self, provider_path):
#         """Append a provider with a given path."""
#         mod_name = 'tackle.providers.' + os.path.basename(provider_path)
#         hooks_init_path = os.path.join(provider_path, 'hooks', '__init__.py')
#         hooks_path = os.path.join(provider_path, 'hooks')
#
#         # If the provider has an __init__.py in the hooks directory, import that
#         # to check if there are any hook types declared there.  If there are, store
#         # those references so that if the hook type is later called, the hook can
#         # then be imported.
#         if os.path.isfile(hooks_init_path):
#             # mod = import_module_from_path(mod_name, hooks_init_path)
#             loader = importlib.machinery.SourceFileLoader(mod_name, hooks_init_path)
#             mod = loader.load_module()
#
#             try:
#                 hook_types = mod.hook_types
#                 p = Provider(path=hooks_path, hook_types=hook_types, name=mod_name)
#                 logger.debug(
#                     f"Importing hook_types {mod.hook_types} from {hooks_path}."
#                 )
#                 self.append(p)
#             except AttributeError:
#                 logger.debug(
#                     f"No `hook_types` declared in {hooks_path}, importing module."
#                 )
#                 # import_with_fallback_install(mod_name, hooks_path)
#         # else:
#         # This pass will import all the modules and when retrieving hook types,
#         # wil search based on all subclasses
#         # self.append(get_provider_from_dir(mod_name, hooks_path))
#         import_with_fallback_install(mod_name, hooks_path, skip_on_error=True)
#
#         p = Provider(path=hooks_path, name=mod_name)
#         self.append(p)
#
#
# def import_with_fallback_install(mod_name, path, skip_on_error: bool = False):
#     """Import a module and on import error, fallback on requirements file."""
#     try:
#         import_hooks_from_dir(mod_name, path, skip_on_error)
#     except ModuleNotFoundError:
#         requirements_path = os.path.join(path, '..', 'requirements.txt')
#         if os.path.isfile(requirements_path):
#             # It is a convention of providers to have a requirements file at the base.
#             # Install the contents if there was an import error
#             subprocess.check_call(
#                 [
#                     sys.executable,
#                     "-m",
#                     "pip",
#                     "install",
#                     "--quiet",
#                     "--disable-pip-version-check",
#                     "-r",
#                     requirements_path,
#                 ]
#             )
#         import_hooks_from_dir(mod_name, path)
#
#
# def import_hooks_from_dir(
#     mod_name,
#     path,
#     skip_on_error: bool = False,
# ):
#     """
#     Import hooks from a directory. This is meant to be used by generically pointing to
#      a hooks directory and importing all the relevant hooks into the context.
#     """
#     potential_hooks = listdir_absolute(path)
#     for f in potential_hooks:
#         if skip_on_error:
#             try:
#                 import_hook_from_path(mod_name, f)
#             except (ModuleNotFoundError, ConfigError):
#                 logger.debug(f"Skipping importing {f}")
#                 continue
#         else:
#             import_hook_from_path(mod_name, f)
#
#
# def import_hook_from_path(
#     mod_name,
#     file_path,
# ):
#     """Import a single hook from a path."""
#     # Maintaining cookiecutter support here as it might have a `hooks` dir.
#     excluded_file_names = ['pre_gen_project', 'post_gen_project', '__pycache__']
#     excluded_file_extensions = ['pyc']
#
#     file_base = os.path.basename(file_path).split('.')
#     if file_base[0] in excluded_file_names:
#         return
#     if file_base[-1] in excluded_file_extensions:
#         return
#
#     if os.path.basename(file_path).split('.')[-1] != "py":
#         # Only import python files
#         return
#
#     loader = importlib.machinery.SourceFileLoader(
#         mod_name + '.' + os.path.basename(file_path).split('.')[0], file_path
#     )
#     loader.load_module()



import os
import re
import sys
import importlib
import logging
from pydantic import BaseModel, Field, ConfigError
from pydantic.main import ModelMetaclass
import subprocess
from typing import TYPE_CHECKING

# TODO: RM after dealing with namespace issue for validators
# https://github.com/robcxyz/tackle-box/issues/43
# import random
# import string

from tackle.utils.paths import listdir_absolute
from tackle.utils.files import read_config_file

if TYPE_CHECKING:
    from tackle.models import Context

logger = logging.getLogger(__name__)


class LazyImportHook(BaseModel):
    """Object to hold hook metadata so that it can be imported only when called."""

    hooks_path: str
    mod_name: str
    provider_name: str
    hook_type: str
    is_public: bool = False

    def wrapped_exec(self, **kwargs):
        kwargs['provider_hooks'].import_with_fallback_install(
            mod_name=self.mod_name,
            path=self.hooks_path,
        )
        return kwargs['provider_hooks'][self.hook_type].wrapped_exec()


class LazyBaseFunction(BaseModel):
    """
    Base function that declarative hooks are derived from and either imported when a
     tackle file is read (by searching in adjacent hooks directory) or on init in local
     providers. Used by jinja extensions and filters.
    """

    function_dict: dict = Field(
        ...,
        description="A dict for the lazy function to be parsed at runtime. Serves as a "
        "carrier for the function's schema until it is compiled with "
        "`create_function_model`.",
    )
    function_fields: list = Field(
        None,
        description="List of fields used to 1, enrich functions without exec method, "
        "and 2, inherit base attributes into methods. Basically a helper.",
    )


def import_from_path(
    context: 'Context',
    provider_path: str,
    hooks_dir_name: str = None,
    skip_on_error: bool = True,
):
    """Append a provider with a given path."""
    # Look for `.hooks` or `hooks` dir
    if hooks_dir_name is None:
        provider_contents = os.listdir(provider_path)
        if '.hooks' in provider_contents:
            hooks_dir_name = '.hooks'
        elif 'hooks' in provider_contents:
            hooks_dir_name = 'hooks'
        else:
            return

    provider_name = os.path.basename(provider_path)
    mod_name = 'tackle.providers.' + provider_name
    hooks_init_path = os.path.join(provider_path, hooks_dir_name, '__init__.py')
    hooks_path = os.path.join(provider_path, hooks_dir_name)

    # If the provider has an __init__.py in the hooks directory, import that
    # to check if there are any hook types declared there.  If there are, store
    # those references so that if the hook type is later called, the hook can
    # then be imported.
    if os.path.isfile(hooks_init_path):
        loader = importlib.machinery.SourceFileLoader(mod_name, hooks_init_path)
        mod = loader.load_module()
        hook_types = getattr(mod, 'hook_types', [])
        for h in hook_types:
            hook = LazyImportHook(
                hooks_path=hooks_path,
                mod_name=mod_name,
                provider_name=provider_name,
                hook_type=h,
            )

            if hook.is_public:
                context.public_hooks[h] = hook
            else:
                context.private_hooks[h] = hook

    # This pass will import all the modules and extract hooks
    import_with_fallback_install(
        context=context, mod_name=mod_name, path=hooks_path, skip_on_error=skip_on_error
    )


def get_native_provider_paths():
    """Get a list of paths to the native providers."""
    providers_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'providers'
    )
    native_providers = [
        os.path.join(providers_path, f)
        for f in os.listdir(providers_path)
        if os.path.isdir(os.path.join(providers_path, f)) and f != '__pycache__'
    ]
    return native_providers


def import_native_providers(context: 'Context'):
    """Iterate through paths and import them."""
    native_provider_paths = get_native_provider_paths()
    for i in native_provider_paths:
        import_from_path(context, i, hooks_dir_name='hooks')


def import_hook_from_path(
    context: 'Context',
    mod_name: str,
    file_path: str,
):
    """Import a single hook from a path."""
    file_split = os.path.basename(file_path).split('.')
    file_base = file_split[0]
    file_extension = file_split[-1]
    # Maintaining cookiecutter support here as it might have a `hooks` dir.
    if file_base in ('pre_gen_project', 'post_gen_project', '__pycache__'):
        return
    if file_extension == 'pyc':
        return
    if file_extension in ('yaml', 'yml'):
        # TODO: Turn this into a parser so that declarative hooks can be imported.
        #  Difference is now that we have access modifiers for hooks where

        # Import declarative hooks
        file_contents = read_config_file(file_path)
        if file_contents is None:
            if context.verbose:
                print(f"Skipping importing {file_path} as the context is empty.")
            return
        for k, v in file_contents.items():
            if re.match(r'^[a-zA-Z0-9\_]*(<\-)$', k):
                hook_type = k[:-2]
                # context.public_context[hook_type] = LazyBaseFunction(
                #     function_dict=v,
                #     # function_fields=[],
                # )
                context.public_hooks[hook_type] = LazyBaseFunction(
                    function_dict=v,
                    # function_fields=[],
                )
            elif re.match(r'^[a-zA-Z0-9\_]*(<\_)$', k):
                hook_type = k[:-2]
                # context.private_context[hook_type] = LazyBaseFunction(
                #     function_dict=v,
                #     # function_fields=[],
                # )
                context.private_hooks[hook_type] = LazyBaseFunction(
                    function_dict=v,
                    # function_fields=[],
                )
        return

    if os.path.basename(file_path).split('.')[-1] != "py":
        # Only import python files
        return

    # TODO: RM after dealing with namespace issue for validators
    # Use a unique RUN_ID to prevent duplicate validator errors
    # https://github.com/robcxyz/tackle-box/issues/43
    # _run_id = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(4))
    # module_name = mod_name + '.hooks.' + file_base[0] + _run_id
    # TODO: Correct version
    # module_name = mod_name + '.hooks.' + file_base[0]
    # TODO: Hack
    try:
        module_name = mod_name + '.hooks.' + file_base[0] + context.run_id
    except AttributeError:
        pass

    loader = importlib.machinery.SourceFileLoader(module_name, file_path)

    if sys.version_info.minor < 10:
        mod = loader.load_module()
    else:
        spec = importlib.util.spec_from_loader(loader.name, loader)
        mod = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = mod
        loader.exec_module(mod)

    for k, v in mod.__dict__.items():
        if (
            isinstance(v, ModelMetaclass)
            and 'hook_type' in v.__fields__
            and k != 'BaseHook'
            and v.__fields__['hook_type'].default is not None
        ):
            # self[v.__fields__['/hook_type'].default] = v
            context.private_hooks[v.__fields__['hook_type'].default] = v


def import_hooks_from_dir(
    context: 'Context',
    mod_name: str,
    path: str,
    skip_on_error: bool = False,
):
    """
    Import hooks from a directory. This is meant to be used by generically pointing to
     a hooks directory and importing all the relevant hooks into the context.
    """
    potential_hooks = listdir_absolute(path)
    for f in potential_hooks:
        if skip_on_error:
            try:
                import_hook_from_path(context, mod_name, f)
            except (ModuleNotFoundError, ConfigError, ImportError):
                logger.debug(f"Skipping importing {f}")
                continue
        else:
            import_hook_from_path(context, mod_name, f)


def import_with_fallback_install(
    context: 'Context',
    mod_name: str,
    path: str,
    skip_on_error: bool = False,
):
    """
    Import a module and on import error, fallback on requirements file and try to
     import again.
    """
    try:
        import_hooks_from_dir(context, mod_name, path, skip_on_error)
    except ModuleNotFoundError:
        requirements_path = os.path.join(path, '..', 'requirements.txt')
        if os.path.isfile(requirements_path):
            # It is a convention of providers to have a requirements file at the base.
            # Install the contents if there was an import error
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--quiet",
                    "--disable-pip-version-check",
                    "-r",
                    requirements_path,
                ]
            )
        import_hooks_from_dir(context, mod_name, path)



from ruamel.yaml import YAML
from tackle import tackle


def test_parser_methods_merge(change_curdir_fixtures):
    yaml = YAML()
    with open('petstore.yaml') as f:
        expected_output = yaml.load(f)

    output = tackle('merge-petstore-compact.yaml')
    assert output == expected_output


def test_parser_methods_try(chdir):
    """Use try which should not have any output"""
    chdir("method-fixtures")
    output = tackle('method-try.yaml')
    assert output == {}


def test_parser_methods_except(chdir):
    """Use try which should not have any output"""
    chdir("method-fixtures")
    output = tackle('method-except.yaml')
    assert output['compact'] == 'foo'
    assert output['str'] == 'foo'
    assert output['dic']['stuff'] == '{{stuff}}'
    assert output['dict_render_block'] == {'stuff': '{{stuff}}'}
    assert output['stuff'] == 'things'
    assert output['listed']['hook_call'][1]['stuff'] == 'things'


def test_parser_methods_when(chdir):
    """Use try which should not have any output"""
    chdir("method-fixtures")
    output = tackle('method-when.yaml')
    assert 'expanded' not in output


def test_parser_methods_else_hooks(chdir):
    """Use try which should not have any output"""
    chdir("method-fixtures")
    output = tackle('method-else.yaml')
    assert output['compact'] == 'foo'
    assert output['str'] == 'foo'
    assert output['dict_render']['stuff'] == '{{stuff}}'
    assert output['str_render_block'] == 'things'
    assert output['dict_render_block'] == {'stuff': '{{stuff}}'}
    assert output['stuff'] == 'things'
    assert output['listed']['hook_call'][1]['stuff'] == 'things'


def test_parser_list_comprehension(chdir):
    """Test that we can do list comprehensions."""
    chdir("method-fixtures")
    output = tackle('list-comprehension.yaml')
    assert len(output['nodes']) == 1


def test_parser_validation_with_try_except(chdir):
    """Test that we can do list comprehensions."""
    chdir("method-fixtures")
    output = tackle('try-validation-except.yaml')
    assert 'p' in output['call']



import os
from ruamel.yaml import YAML
import subprocess

from tackle.models import BaseHook, Field


class K8sCurrentContextHook(BaseHook):
    """Hook for getting current kubeconfig context."""

    hook_type: str = 'k8s_current_context'

    def exec(self) -> str:
        kubeconfig_locations = os.environ.get("KUBECONFIG", None)

        if kubeconfig_locations is None:
            raise Exception("KUBECONFIG not set, exiting...")
        yaml = YAML()
        with open(kubeconfig_locations.split(':')[0]) as f:
            context = yaml.load(f)

        return context['current-context']


class K8sContextListHook(BaseHook):
    """Hook for listing kubeconfig contexts."""

    hook_type: str = 'k8s_context_list'

    def exec(self) -> list:
        kubeconfig_locations = os.environ.get("KUBECONFIG", None)

        if kubeconfig_locations is None:
            raise Exception("KUBECONFIG not set, exiting...")

        output_contexts = []
        yaml = YAML()
        for c in kubeconfig_locations.split(':'):
            with open(c) as f:
                context = yaml.load(f)

            for i in context['contexts']:

                if i['name'] not in output_contexts:
                    output_contexts.append(i['name'])
        return output_contexts


class K8sContextMapHook(BaseHook):
    """Hook for return a map of the kubeconfig context details."""

    hook_type: str = 'k8s_context_map'

    def exec(self) -> dict:
        kubeconfig_locations = os.environ.get("KUBECONFIG", None)

        if kubeconfig_locations is None:
            raise Exception("KUBECONFIG not set, exiting...")

        output_contexts = {}
        yaml = YAML()
        for c in kubeconfig_locations.split(':'):
            with open(c) as f:
                context = yaml.load(f)

            for i in context['contexts']:
                if i['name'] not in output_contexts.keys():
                    output_contexts[i['name']] = context
        return output_contexts


class K8sUseContextHook(BaseHook):
    """Hook for using a kube context."""

    hook_type: str = 'k8s_use_context'
    context: str = Field(..., description="The context to use.")

    def exec(self):
        subprocess.run(
            ["kubectl", "config", "use-context", self.context], stdout=subprocess.PIPE
        )



from typing import Any
from tackle.models import BaseHook, Field
from tackle.parser import parse_tmp_context


class RunHookHook(BaseHook):
    """Hook to run other hooks dynamically."""

    hook_type: str = 'run_hook'

    hook_name: str = Field(..., description="The name of the hook.")
    hook_dict: dict = Field(None, description="A dict of keys to use with the hook.")

    args: list = ['hook_name', 'hook_dict']
    kwargs: str = 'hook_dict'

    def exec(self) -> Any:
        element = {'tmp': {self.key_path[-1]: self.hook_name, **self.hook_dict}}

        output = parse_tmp_context(self, element, existing_context=self.public_context)
        return output['tmp']



"""Tests dict input objects for `tackle.providers.system.hooks.symlink` module."""
from tackle.main import tackle
import pytest
import os


@pytest.fixture()
def clean_up_outputs():
    """Clean up."""
    yield
    os.unlink('dst.yaml')


def test_provider_system_hook_symlink_overwrite(change_dir, clean_up_outputs):
    output = tackle('tackle.yaml', no_input=True)

    assert output['read'] == {'foo': 'bar'}
    assert output['read_dir'] == {'foo': 'bar'}
    assert output['read_dir2'] == {'foo': 'bar'}



# from tackle import tackle
#
# # TODO: https://github.com/robcxyz/tackle-box/issues/46
#
#
# def test_provider_select_list_this(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.select.prompt', return_value={"tmp": "things"}
#     )
#     output = tackle('test.yaml')
#     assert output['selection'] == 'things'
#
#
# def test_provider_select_list(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.select.prompt', return_value={"tmp": "things"}
#     )
#     output = tackle('list.yaml')
#     assert output['selection'] == 'things'
#
#
# def test_provider_select_list_index(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.select.prompt', return_value={"tmp": "things"}
#     )
#     output = tackle('list_index.yaml')
#     assert output['selection'] == 1
#
#
# def test_provider_select_map(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.select.prompt',
#         return_value={"tmp": "I do things"},
#     )
#     output = tackle('map.yaml')
#     assert output['selection'] == 'things'
#
#
# def test_provider_select_map_index(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.select.prompt',
#         return_value={"tmp": "I do things"},
#     )
#     output = tackle('map_index.yaml')
#     assert output['selection'] == 1


# def test_provider_select_no_msg(change_dir, mocker):
#     mocker.patch(
#         'tackle.providers.prompts.hooks.select.prompt', return_value={"tmp": "things"}
#     )
#     output = tackle('no-msg.yaml')
#     assert output['selection'] == 'things'



import sys
from InquirerPy import prompt
from InquirerPy.base.control import Choice

from typing import Union, List, Dict

from tackle.models import BaseHook, Field
from tackle.utils.dicts import get_readable_key_path


class InquirerCheckboxHook(BaseHook):
    """
    Hook for PyInquirer `checkbox` type prompts. Allows the user to multi-select from a
     list of choices and outputs a list. Takes in three forms of `choices` inputs. A
     list of string, a list of maps with all keys having a `name` field per the original
     [spec](https://github.com/CITGuru/PyInquirer/blob/master/examples/checkbox.py), or
     list of maps with the key as the question, the value as the output.
    """

    hook_type: str = 'checkbox'

    message: str = Field(None, description="String message to show when prompting.")
    # default: Any = Field([], description="Default for the return value")
    choices: Union[List[str], List[Dict]] = Field(
        ...,
        description="Either a list of strings or dictionary .",
        render_by_default=True,
    )
    checked: bool = Field(
        False, description="Boolean if the default choices should all be checked."
    )
    index: bool = Field(
        False, description="Boolean to return the index instead of the answer"
    )
    fuzzy: bool = Field(
        False,
        description="Allow fuzzy search like fzf - [docs](https://inquirerpy.readthedocs.io/en/latest/pages/prompts/fuzzy.html)",
    )

    args: list = ['message']
    _docs_order: int = 2

    def exec(self) -> list:
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

        choices_type = None
        for i, v in enumerate(self.choices):
            if i != 0 and type(v) != choices_type:
                raise ValueError("All items need to be of the same type")
            choices_type = type(v)

        if choices_type == str:
            if self.no_input:
                if self.checked:
                    return self.choices
                return []

            choice_list = self.choices.copy()
            self.choices = [
                {'name': x, 'value': x} if isinstance(x, str) else x
                for x in self.choices
            ]

            answer = self._run_prompt()
            if self.index:
                output = []
                for i, v in enumerate(choice_list):
                    if v in answer:
                        output.append(i)
                return output
            else:
                return answer

        elif choices_type == dict:
            # Check if the input is in the normal form of how pyinquirer takes choices:
            # choices = [{'name': 'stuff'}, {'name': 'things','checked':True}]
            # https://github.com/kazhala/InquirerPy/blob/master/examples/checkbox.py
            normal = True
            for i in self.choices:
                if 'name' not in i or 'value' not in i:
                    normal = False
                    break

            if normal:
                if self.no_input:
                    if self.checked:
                        return [i['name'] for i in self.choices]
                    return []

                if self.index:
                    # TODO: Fix this - low priority
                    raise Exception("Can't index checkbox calls in normal form.")
                return self._run_prompt()

            # Otherwise we expect to reindex the key as the output per this:
            # choices = ['How much stuff?': 'stuff', 'How many things?': 'things']
            choices = []
            if self.no_input:
                if self.checked:
                    for i in self.choices:
                        choices.append(list(i.values())[0])
                    return choices
                return []

            for i in self.choices:
                choices.append(list(i.keys())[0])

            # Fixing to the expected input choices {'name': 'stuff', 'name': ...}
            choices_map = self.choices.copy()
            self.choices = [
                {'name': x, 'value': x} if isinstance(x, str) else x for x in choices
            ]

            answer = self._run_prompt()

            # This is for reindexing the output
            output = []
            for i, v in enumerate(choices_map):
                val = list(v.keys())[0]
                if val in answer:
                    if self.index:
                        output.append(i)
                    else:
                        output.append(v[val])
            return output

    def _run_prompt(self):
        if self.checked:
            choices = [
                Choice(**i, enabled=True) if isinstance(i, dict) else i
                for i in self.choices
            ]
        else:
            choices = self.choices

        question = {
            'type': 'checkbox' if not self.fuzzy else 'fuzzy',
            'name': 'tmp',
            'message': self.message,
            'choices': choices,
        }
        try:
            response = prompt([question])
        except KeyboardInterrupt:
            print("Exiting...")
            sys.exit(0)
        return response['tmp']



from tackle import tackle


###############
# Default hooks
###############
def test_function_default_hook_no_context(change_curdir_fixtures):
    """Validate that we can run a default hook."""
    output = tackle('cli-default-hook-no-context.yaml')
    assert output['p'] == 'things'


def test_function_default_hook_no_context_kwargs(change_curdir_fixtures):
    """Validate that we can run a default hook with a kwarg."""
    output = tackle('cli-default-hook-no-context.yaml', stuff='bar')
    assert output['p'] == 'bar'
    assert output['b']


def test_function_default_hook_no_context_flags(change_curdir_fixtures):
    """
    Validate that flags are interpreted properly. In this case the default is true so
    when setting the flag, it should be false.
    """
    output = tackle('cli-default-hook-no-context.yaml', global_flags=['things'])
    assert isinstance(output['b'], bool)
    assert not output['b']


def test_function_default_hook_context(change_curdir_fixtures):
    """Test that outer context is additionally parsed with the default hook."""
    output = tackle('cli-default-hook-context.yaml')
    assert output['p'] == 'things'
    assert output['foo'] == 'bar'


def test_function_default_hook_no_context_method_call(change_curdir_fixtures):
    """Validate that we can run a default hook."""
    output = tackle('cli-default-hook-no-context.yaml', 'do')
    assert output['d'] == 'baz'


def test_function_default_hook_no_context_method_call_args(change_curdir_fixtures):
    """Validate that we can run a default hook."""
    output = tackle('cli-default-hook-no-context.yaml', 'do', 'bizz')
    assert output['d'] == 'bizz'


def test_function_default_hook_embedded(change_curdir_fixtures):
    """Validate that we can run a default hook embedded methods."""
    output = tackle('cli-default-hook-embedded.yaml', 'do', 'stuff', 'things')
    assert output['t'] == 'bar'


def test_function_default_hook_embedded_kwargs(change_curdir_fixtures):
    """Validate that we can run a default hook embedded methods with kwargs."""
    output = tackle(
        'cli-default-hook-embedded.yaml', 'do', 'stuff', 'things', foo='bing'
    )
    assert output['t'] == 'bing'


def test_function_default_hook_embedded_kwargs_full(change_curdir_fixtures):
    """Validate that we can run a default hook embedded methods with kwargs schema."""
    output = tackle('cli-default-hook-embedded.yaml', 'do', 'stuff', foo_full='bing')
    assert output['t'] == 'bing'


#############
# Non-default
#############
def test_function_cli_hook_arg(change_curdir_fixtures):
    """Validate that we can run a default hook with an arg."""
    output = tackle('cli-hook-no-context.yaml', 'run')
    assert output['t'] == 'things'
    assert output['s']


def test_function_cli_hook_arg_args(change_curdir_fixtures):
    """Validate that we can run a default hook with an arg."""
    output = tackle('cli-hook-no-context.yaml', 'run', 'do', 'bazzz')
    assert output['d'] == 'bazzz'
    assert output['t'] == 'things'
    assert output['s']


def test_function_cli_hook_arg_kwargs(change_curdir_fixtures):
    """Validate that we can run a default hook with an arg."""
    output = tackle('cli-hook-no-context.yaml', 'run', stuff='bazzz')
    assert output['t'] == 'bazzz'
    assert output['s']


def test_function_cli_hook_arg_flags(change_curdir_fixtures):
    """Validate that we can run a default hook with an arg."""
    output = tackle('cli-hook-no-context.yaml', 'run', global_flags=['things'])
    assert not output['s']


def test_function_hook_embedded_kwargs(change_curdir_fixtures):
    """Validate that we can run a default hook embedded methods with kwargs."""
    output = tackle(
        'cli-default-hook-embedded.yaml', 'run', 'do', 'stuff', 'things', foo='bing'
    )
    assert output['t'] == 'bing'



import webbrowser

from tackle.models import BaseHook, Field


class WebBrowserHook(BaseHook):
    """
    Open a web browser. Wraps python's
     [webbrowser](https://docs.python.org/3/library/webbrowser.html#browser-controller-objects)
     module.
    """

    hook_type: str = 'webbrowser'
    # fmt: off
    url: str = Field(..., description="String url to open in browser.")
    new: int = Field(0, description="If new is 1, a new browser window is opened if possible. If new is 2, a new browser page (“tab”) is opened if possible.")
    autoraise: bool = Field(True, description="If autoraise is True, the window is raised if possible (note that under many window managers this will occur regardless of the setting of this variable).")
    # fmt: on

    args: list = ['url']

    def exec(self):
        webbrowser.open(self.url, new=2)



"""Tests dict input objects for `tackle.providers.github.hooks` modules."""
import os
from tackle.main import tackle
import pytest
from tackle.utils.paths import rmtree


@pytest.fixture()
def clean_outputs():
    """Clean the outputs."""
    yield
    paths = ['services', 'foo', 'tackle-pypackage']
    for p in paths:
        if os.path.isdir(p):
            rmtree(p)


def test_provider_git_hooks_repos(change_dir, clean_outputs):
    """Verify hook."""
    tackle('repo.yaml')
    assert os.path.isdir('tackle-pypackage')


def test_provider_git_hooks_meta_flat(change_dir, clean_outputs):
    """Verify hook."""
    tackle('meta-flat.yaml', no_input=True)
    assert os.path.isdir('foo')


def test_provider_git_hooks_meta(change_dir, clean_outputs):
    """Verify hook."""
    tackle('meta.yaml', no_input=True)
    assert os.path.isdir(os.path.join('foo', 'bar', 'thing', 'docs'))
    assert os.path.isdir(os.path.join('foo', 'bar', 'gh', 'docs'))
    assert os.path.isdir(os.path.join('foo', 'bar', 'gh-branch', 'docs'))
    assert os.path.isdir(os.path.join('foo', 'bar', 'main', 'docs'))
    assert os.path.isdir(os.path.join('services', 'users', 'docs'))
    assert os.path.isdir(os.path.join('services', 'thing', 'docs'))



from typing import Union, Optional, Any

from tackle import BaseHook, Field
from tackle.utils.dicts import (
    encode_key_path,
    nested_set,
    get_target_and_key,
    nested_get,
)


class DictUpdateHook(BaseHook, smart_union=True):
    """
    Hook for updating dict objects with values, appending list values, or overwriting
     string / int / float values.
    """

    hook_type: str = 'update'

    # fmt: off
    src: Union[dict, str, list] = Field(
        description="A dict to update and output the result or a str with separators "
                    "or list for a key path to the item update within the context.")
    input: Any = Field(description="The value to update the input `src`.")
    sep: str = Field('/', description="For string src's, a separator for key path.")
    # fmt: on

    args: list = ['src', 'input']

    def exec(self) -> Optional[dict]:
        if isinstance(self.src, (str, list)):
            self.src = encode_key_path(self.src, self.sep)
        if isinstance(self.src, list):
            target_context, set_key_path = get_target_and_key(self, key_path=self.src)

            src = nested_get(
                element=target_context,
                keys=set_key_path,
            )
            if isinstance(src, dict):
                src.update(self.input)
            elif isinstance(src, list):
                src.append(self.input)
            else:
                # This overwrites the input
                nested_set(
                    element=target_context,
                    keys=set_key_path,
                    value=self.input,
                )
        else:
            self.src.update(self.input)
            return self.src



"""Test tackle.utils.paths."""
import pytest
from tackle.utils.paths import is_file, is_repo_url

FILE_FIXTURES = [
    ("file.yaml", True),
    ("file.yml", True),
    ("/foo.yaml/.file.yml", True),
    (".file.yml", True),
    ("file.json", True),
    ("file.hcl", False),
]


@pytest.mark.parametrize("file,is_file_result", FILE_FIXTURES)
def test_is_file(file, is_file_result):
    """Validate is_file regex."""
    assert is_file(file) == is_file_result


REPO_FIXTURES = [
    ("https://github.com/robcxyz/tackle-demo", True),
    ("github.com/robcxyz/tackle-demo", True),
    ("gitlab.com/robcxyz/tackle-demo", True),
    ("gitlab.com/robcxyz/tackle-demo/some/things", False),
    ("stuffhub.com/robcxyz/tackle-demo", False),
    ("robcxyz/tackle-demo", True),
    ("tests/utils", False),
    ("tests/utils/foo", False),
]


@pytest.mark.parametrize("repo,is_repo_result", REPO_FIXTURES)
def test_is_repo_url(repo, is_repo_result, change_dir_base):
    """Validate is_repo_url regex."""
    assert is_repo_url(repo) == is_repo_result



import os
import pytest

from tackle.main import tackle


@pytest.fixture()
def set_kubeconfig():
    # fmt: off
    os.environ["KUBECONFIG"] = ':'.join([os.path.join(os.path.abspath(os.path.dirname(__file__)), f'kubeconfig{i}.yaml') for i in ['', '2']])
    # fmt: on


def test_provider_k8s_context(change_dir, set_kubeconfig):
    output = tackle('context.yaml', no_input=True)
    assert 'federal-context' in output['context_map']
    assert output['context'] == 'federal-context'



from tackle.main import tackle


def test_provider_system_hook_random_string(change_dir):
    output = tackle('random.yaml')

    assert len(output['random_hex']) == 8
    assert len(output['random_hex_arg']) == 4
    assert len(output['random_string']) == 8
    assert len(output['random_string_arg']) == 6
    assert output['random_string_upper_arg'].isupper()



from tackle import tackle


def test_provider_assert(change_dir):
    """Check assertions."""
    output = tackle('assert.yaml')
    assert 'assert' in output
    assert 'no_exist' not in output
    assert output['assert_lhs']



from tackle import tackle


def test_hook_generate_update_document(change_dir):
    """Check that we can update a section in a document and assert the proper output."""
    output = tackle("basic.yaml")
    with open('expected-output.md') as f:
        expected_output = f.read()
    with open('file.md') as f:
        file = f.read()

    assert output
    assert file == expected_output


def test_hook_generate_update_document_multi_line(change_dir):
    """Check multi-line capability in arg."""
    output = tackle("multi-line.yaml", "update_readme")

    with open('multi-line.md') as f:
        file = f.read()

    assert '| Stuff | A thing |' in file
    assert output



"""Tests dict input objects for `tackle.providers.rich.hooks.table` module."""
from tackle.main import tackle


def test_provider_system_hook_rich_table(change_dir):
    output = tackle('table.yaml', no_input=True)
    assert 'table' in output


def test_provider_system_hook_rich_table_split(change_dir):
    output = tackle('table_split.yaml', no_input=True)
    assert 'table_split' in output


def test_provider_console_markdown(change_dir):
    output = tackle('markdown.yaml')
    assert output['md']



"""Example."""
from tackle.models import BaseHook


class StuffHook(BaseHook):
    """Do stuff and things."""

    hook_type: str = 'stuff'
    thing: str

    def exec(self):
        print(self.thing)
        return self.thing



import pytest

from tackle import exceptions, main


def test_should_raise_error_if_repo_does_not_exist(chdir):
    """Tackle invocation with non-exist repository should raise error."""
    chdir('/')
    with pytest.raises(exceptions.UnknownSourceException) as info:
        main.tackle('definitely-not-a-valid-repo-dir')

    assert 'definitely-not-a-valid-repo-dir' in info.value.message


def test_should_raise_error_with_calling_file(change_curdir_fixtures):
    """The correct file with error should come up in error message."""
    with pytest.raises(exceptions.UnknownSourceException) as info:
        main.tackle('calling-tackle.yaml', verbose=True)
    # This assumes there is a tackle file in a parent directory
    assert 'Error parsing input_file=' in info.value.message


FIXTURES = [
    # (
    #     'unknown-argument-extra.yaml',
    #     exceptions.UnknownArgumentException,
    #     'fixtures/unknown-argument.yaml',
    # ),
    (
        'unknown-argument.yaml',
        exceptions.UnknownArgumentException,
        'fixtures/unknown-argument.yaml',
    ),
    (
        'unknown-variable.yaml',
        exceptions.UnknownTemplateVariableException,
        'fixtures/unknown-variable.yaml',
    ),
    (
        'unknown-variable-call.yaml',
        exceptions.UnknownTemplateVariableException,
        'fixtures/unknown-variable.yaml',
    ),
    (
        'unknown-variable-call-tackle.yaml',
        exceptions.UnknownTemplateVariableException,
        'fixtures/tackle.yaml',
    ),
    ('bad-extension.yaml', exceptions.UnknownTemplateVariableException, 'foobar'),
    (
        'unknown-named-argument.yaml',
        exceptions.UnknownInputArgumentException,
        'unknown-named-argument.yaml',
    ),
    (
        'missing-quote.yaml',
        exceptions.MalformedTemplateVariableException,
        'missing-quote.yaml',
    ),
]


@pytest.mark.parametrize("fixture,exception,msg", FIXTURES)
def test_should_raise_error_fixtures(change_curdir_fixtures, fixture, exception, msg):
    """Assert right exception is raised with helper `msg`."""
    with pytest.raises(exception) as e:
        main.tackle(fixture, verbose=True)

    assert msg in e.value.message



"""Main tests."""
import os
import sys

import pytest

from tackle import tackle
from tackle.cli import main
from tackle import exceptions


def test_main_cli_call_mock(mocker):
    """Check the main function runs properly."""
    mock = mocker.patch("tackle.main.update_source")
    main("stuff")
    assert mock.called


def test_main_cli_call_empty(change_curdir_fixtures, mocker):
    """
    Check that when no arg is given that we find the closes tackle file which
    could be in the parent directory.
    """
    mock = mocker.patch("tackle.main.update_source")
    main([])
    assert mock.called
    local_tackle = os.path.join(os.path.abspath('.'), '.tackle.yaml')

    if sys.version_info.minor > 7:
        assert mock.call_args.args[0].input_string == local_tackle
    # test was failing in 3.7/6
    else:
        assert mock.call_args[0][0].input_string == local_tackle


def test_main_cli_call_empty_no_parent_tackle_raises(chdir):
    """
    Check that when no arg is given that we find the closes tackle file which
    could be in the parent directory.
    """
    if os.name != 'nt':
        chdir('/')
    else:
        chdir('\\')
    with pytest.raises(exceptions.NoInputOrParentTackleException):
        tackle()


def test_main_input_dict(change_curdir_fixtures):
    """
    Test bringing in an input dict along with a target which should be interpreted as
    overriding keys to the target.
    """
    input_dict = {
        'this': 1,
        'that': 2,
        'stuff': 'things',  # Missing key
        'foo': 2,  # Non-hook key
    }

    output = tackle('dict-input.yaml', **input_dict)
    assert output['this'] == 1


def test_main_from_cli_input_dict(change_curdir_fixtures, capsys):
    """Test same as above but from command line."""
    main(["dict-input.yaml", "--this", "1", "--that", "2", "--print"])
    assert 'this' in capsys.readouterr().out


def test_main_overrides_str(change_curdir_fixtures):
    """Test that we can override inputs."""
    o = tackle("dict-input.yaml", override="dict-input-overrides.yaml")
    # Should normally throw error with prompt
    assert o['this'] == "stuff"
    # Again, should raise error w/o override
    main(["dict-input.yaml", "--override", "dict-input-overrides.yaml"])


def test_main_overrides_str_not_found_error(change_curdir_fixtures):
    """Test that we get error on ."""
    with pytest.raises(exceptions.UnknownInputArgumentException):
        tackle("dict-input.yaml", override="not-exists")



import os
from ruamel.yaml import YAML

import pytest
from tackle.main import tackle


@pytest.fixture()
def clean_outputs(change_dir):
    """Remove all the files prefixed with output before and after test."""
    files = [f for f in os.listdir() if f.split('.')[0].startswith('output')]
    for f in files:
        os.remove(f)
    yield
    files = [f for f in os.listdir() if f.split('.')[0].startswith('output')]
    for f in files:
        os.remove(f)


def test_provider_system_hook_yaml_read(change_dir, clean_outputs):
    read = tackle('read.yaml', no_input=True)

    assert read['stuff'] == 'things'


def test_provider_system_hook_yaml_write(change_dir, clean_outputs):
    tackle('write.yaml', no_input=True)
    yaml = YAML()
    with open('output.yaml', 'r') as f:
        written = yaml.load(f)
    assert written == {'stuff': 'things'}


# TODO: When in place yaml hooks are a thing
def test_provider_system_hook_yaml_update(change_dir, clean_outputs):
    tackle('update.yaml', no_input=True)

    yaml = YAML()
    with open('output.yaml', 'r') as f:
        output = yaml.load(f)

    assert output['stuff'] == {'things': {'cats': 'scratch'}}


def test_provider_system_hook_yaml_remove_str(change_dir, clean_outputs):
    tackle('remove_str.yaml', no_input=True)

    yaml = YAML()
    with open('output.yaml', 'r') as f:
        output = yaml.load(f)

    assert output == ['stuff', 'things']


def test_provider_system_hook_yaml_remove_list(change_dir, clean_outputs):
    tackle('remove_list.yaml', no_input=True)

    yaml = YAML()
    with open('output.yaml', 'r') as f:
        output = yaml.load(f)

    assert output == ['stuff', 'things']


def test_provider_system_hook_yaml_filter(change_dir, clean_outputs):
    output = tackle('filter.yaml', no_input=True)

    assert 'stuff' not in output['things']


def test_provider_system_hook_yaml_update_in_place(change_dir, clean_outputs):
    tackle('update_in_place.yaml', no_input=True)

    yaml = YAML()
    with open('output_update_in_place.yaml', 'r') as f:
        output = yaml.load(f)

    assert output['dev']['stuff'] == 'things'


def test_provider_system_hook_yaml_merge_in_place(change_dir, clean_outputs):
    tackle('merge_in_place.yaml', no_input=True)

    yaml = YAML()
    with open('output_merge_in_place.yaml', 'r') as f:
        output = yaml.load(f)

    assert output['dev']['stuff'] == 'things'


def test_provider_system_hook_yaml_append(change_dir, clean_outputs):
    output = tackle('append.yaml', no_input=True)
    assert output['append_dict'] == {'things': ['dogs', 'cats', 'bar', 'baz']}


def test_yaml_yamlify(change_dir, clean_outputs):
    output = tackle('yamlencode.yaml', no_input=True)
    assert isinstance(output['out'], str)
    assert 'stuff: things' in output['out']


def test_yaml_yamldecode(change_dir, clean_outputs):
    output = tackle('yamldecode.yaml', no_input=True)
    assert isinstance(output['out'], dict)
    assert output['out']['stuff'] == 'things'



from tackle.exceptions import TackleHookCallException


class UndefinedVariableInTemplate(TackleHookCallException):
    """
    Exception for out-of-scope variables.

    Raised when a template uses a variable which is not defined in the
    context.
    """


class GenerateHookTemplateNotFound(TackleHookCallException):
    """
    Exception when template is not found.

    Raised when can't find the template to render.
    """



"""Terraform hooks."""
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os
import zipfile
from pydantic import Field

from tackle.models import BaseHook

logger = logging.getLogger(__name__)


class ZipHook(BaseHook):
    """Hook to zip a file or directory."""

    hook_type: str = 'zipfile'

    input: str = Field(..., description="Input path")
    output: str = Field(..., description="Output path")

    def exec(self):
        if os.path.isdir(self.input):
            zf = zipfile.ZipFile(self.output, "w", zipfile.ZIP_DEFLATED)
            for dirname, subdirs, files in os.walk(self.input):
                zf.write(dirname)
                for filename in files:
                    zf.write(os.path.join(dirname, filename))
            zf.close()
        elif os.path.isfile(self.input):
            zipfile.ZipFile(self.output, "w").write(self.input)
        else:
            raise FileNotFoundError(f"Can't find input path {self.input}.")
        return self.output


class UnzipHook(BaseHook):
    """Hook to unzip a file."""

    hook_type: str = 'unzipfile'

    input: str = Field(..., description="Input path")
    output: str = Field(".", description="Output path, default to current directory")

    def exec(self):
        if os.path.isfile(self.input):
            with zipfile.ZipFile(self.input, 'r') as zip_ref:
                zip_ref.extractall(self.output)
        else:
            raise FileNotFoundError(f"Can't find input path {self.input}.")
        return self.output



import sys
import pytest

from tackle import tackle
from tackle.providers.tackle.hooks.provider_docs import hook_type_to_string

from typing import Dict, List, Union, Any

TYPE_FIXTURES = [
    (dict, "dict"),
    # (Dict, "dict"),
    (list, "list"),
    # (List, "list"),
    (Union[dict, list], "union"),
    (Any, "any"),
]


@pytest.mark.parametrize("type_,expected_output", TYPE_FIXTURES)
def test_hook_type_to_string(type_, expected_output):
    output = hook_type_to_string(type_)
    assert output == expected_output


def test_provider_tackle_provider_docs(change_dir):
    """Run the provider docs."""
    if sys.version_info.minor <= 6:
        return

    output = tackle('docs.yaml')
    assert (
        len(
            [
                i
                for i in output['render_context']['hooks']
                if i['hook_type'] == 'provider_docs'
            ]
        )
        == 1
    )



"""Environment variable hooks."""
import os
import logging
import platform

from tackle.models import BaseHook, Field

logger = logging.getLogger(__name__)


class GetEnvHook(BaseHook):
    """Hook for getting environment variables."""

    hook_type: str = 'get_env'
    environment_variable: str = Field(
        None,
        description="Dict for setting and string for getting environment variables",
    )
    fallback: str = Field(None, description="A fallback for getting.")

    args: list = ['environment_variable', 'fallback']

    def exec(self):
        """Set env vars."""
        if platform.system() == 'Windows':
            from tackle.exceptions import ContributionNeededException

            raise ContributionNeededException()
            # robcxyz - Not working on windows support but would love PRs
        else:
            return os.getenv(self.environment_variable, self.fallback)


class EnvironmentVariableHook(BaseHook):
    """Hook for setting environment variables."""

    hook_type: str = 'set_env'
    environment_variable: str = Field(
        ..., description="The name of the environment variable to set."
    )
    value: str = Field(None, description="The value to set it.")

    args: list = ['environment_variable', 'value']

    def exec(self):
        """Get or set env vars."""
        if platform.system() == 'Windows':
            from tackle.exceptions import ContributionNeededException

            raise ContributionNeededException()
            # robcxyz - Not working on windows support but would love PRs
        else:
            # TODO: Implement parsing for strings to set -> export THING=stuff
            os.environ[self.environment_variable] = self.value
            return self.value


class ExportHook(BaseHook):
    """Hook for setting environment variables that returns None."""

    hook_type: str = 'export'
    environment_variable: str = Field(
        ..., description="The name of the environment variable to set."
    )
    value: str = Field(None, description="The value to set it.")

    args: list = ['environment_variable', 'value']

    # TODO: Implement parsing for strings to set -> <: export THING=stuff
    def exec(self):
        """Set env vars."""
        if platform.system() == 'Windows':
            from tackle.exceptions import ContributionNeededException

            raise ContributionNeededException()
            # robcxyz - Not working on windows support but would love PRs
        else:
            os.environ[self.environment_variable] = self.value
        return


class UnsetHook(BaseHook):
    """Hook for unsetting environment variables."""

    hook_type: str = 'unset'
    environment_variable: str = Field(
        ..., description="The name of the environment variable to set."
    )

    args: list = ['environment_variable']

    def exec(self):
        """Set env vars."""
        if platform.system() == 'Windows':
            from tackle.exceptions import ContributionNeededException

            raise ContributionNeededException()
            # robcxyz - Not working on windows support but would love PRs
        else:
            os.unsetenv(self.environment_variable)
        return



import os
import json
import logging
from typing import Union, Any
from pydantic import BaseModel

import requests
from requests.auth import HTTPBasicAuth

from tackle.models import BaseHook, Field

logger = logging.getLogger(__name__)


def exit_none_200(r: requests.Response, no_exit: bool, url: str):
    """Exit if the status code is not 2xx."""
    if not (r.status_code - 200) < 100 and not no_exit:
        raise requests.exceptions.HTTPError(
            f"Error sending request to {url}, got '{r.status_code}' status code."
        )


def process_content(r: requests.Response, encoding):
    """Output the content based on the header."""
    if 'application/json' in r.headers['Content-Type']:
        return json.loads(r.content)
    elif 'text/plain' in r.headers['Content-Type']:
        return r.content.decode(encoding=encoding)


class AuthMixin(BaseModel):
    """Authorization for web requesst."""

    username: str = None
    password: str = None

    def auth(self):
        if self.username is not None:
            return HTTPBasicAuth(username=self.username, password=self.password)
        return None


class RequestsGetHook(BaseHook, AuthMixin):
    """
    Hook for Requests 'get' type prompts.
    [Link](https://docs.python-requests.org/en/latest/api/#requests.get)
    """

    hook_type: str = 'http_get'

    # fmt: off
    url: str = Field(..., description="URL for the new request object.")
    kwargs: Union[str, dict] = Field(
        {}, description="Optional arguments that request takes.",
        render_by_default=True
    )
    params: dict = Field(None,
                         description="Dictionary, list of tuples or bytes to send in the query string for the Request.")
    no_exit: bool = Field(False, description="Whether to exit on non-200 response.")
    encoding: str = Field('utf-8',
                          description="For text/plain type return values, the encoding of the type.")
    # fmt: on

    args: list = [
        'url',
        'kwargs',
        'params',
    ]

    def exec(self) -> dict:
        r = requests.get(self.url, params=self.params, auth=self.auth(), **self.kwargs)
        exit_none_200(r, self.no_exit, self.url)
        return process_content(r, encoding=self.encoding)


class RequestsPostHook(BaseHook, AuthMixin):
    """
    Hook for Requests 'post' type prompts.
    [Link](https://docs.python-requests.org/en/latest/api/#requests.post)
    """

    hook_type: str = 'http_post'

    # fmt: off
    url: str = Field(..., description="URL for the new request object.")
    kwargs: Union[str, dict] = Field(
        {}, description="Optional arguments that request takes.",
        render_by_default=True
    )
    # Requests API docs call for additional functionality not for yaml
    data: Any = Field(
        None,
        description="Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request.",
        render_by_default=True
    )
    # TODO: Fix alias so input_json -> json
    #  https://github.com/robcxyz/tackle-box/issues/80
    input_json: dict = Field(
        False,
        description="Whether to exit on non-200 response.",
        render_by_default=True
    )
    no_exit: bool = Field(False, description="Whether to exit on non-200 response.")
    # fmt: on

    args: list = ['url', 'data', 'kwargs']

    def exec(self) -> dict:
        if isinstance(self.data, str):
            if not os.path.exists(self.data):
                raise FileNotFoundError(
                    f"For the requests patch method, data fields must be a "
                    f"reference to a file path, not {self.data}."
                )

        r = requests.post(self.url, data=self.data, json=self.input_json,
                          auth=self.auth(), **self.kwargs)
        exit_none_200(r, self.no_exit, self.url)

        return r.json()


class RequestsPutHook(BaseHook, AuthMixin):
    """
    Hook for Requests 'put' type prompts.
    [Link](https://docs.python-requests.org/en/latest/api/#requests.put)
    """

    hook_type: str = 'http_put'

    # fmt: on
    url: str = Field(..., description="URL for the new request object.")
    kwargs: Union[str, dict] = Field(
        {}, description="Optional arguments that request takes.",
        render_by_default=True
    )
    data: Any = Field(
        None,
        description="Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request.",
    )
    # TODO: Fix alias so input_json -> json
    #  https://github.com/robcxyz/tackle-box/issues/80
    input_json: dict = Field(
        None,
        description="Json data to send in the body of the Request.",
        render_by_default=True
    )
    no_exit: bool = Field(False, description="Whether to exit on non-200 response.")
    # fmt: off

    args: list = ['url', 'data', 'kwargs']

    def exec(self):
        r = requests.put(self.url, data=self.data, json=self.input_json,
                         auth=self.auth(), **self.kwargs)
        exit_none_200(r, self.no_exit, self.url)

        return r.json()


class RequestsPatchHook(BaseHook, AuthMixin):
    """
    Hook for Requests 'patch' type prompts.
    [Link](https://docs.python-requests.org/en/latest/api/#requests.patch)
    """

    hook_type: str = 'http_patch'

    # fmt: off
    url: str = Field(..., description="URL for the new request object.")
    kwargs: Union[str, dict] = Field(
        {}, description="Optional arguments that request takes.",
        render_by_default=True
    )
    data: Any = Field(
        None,
        description="Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request.",
    )
    # TODO: Fix alias so input_json -> json
    #  https://github.com/robcxyz/tackle-box/issues/80
    input_json: dict = Field(
        None,
        description="Json data to send in the body of the Request.",
        render_by_default=True,
    )
    no_exit: bool = Field(False, description="Whether to exit on non-200 response.")
    # fmt: off

    args: list = ['url', 'data', 'kwargs']

    def exec(self):
        r = requests.patch(
            self.url, data=self.data, json=self.input_json, auth=self.auth(),
            **self.kwargs
        )
        exit_none_200(r, self.no_exit, self.url)

        return r.json()


class RequestsDeleteHook(BaseHook, AuthMixin):
    """
    Hook for Requests 'delete' type prompts.
    [Link](https://docs.python-requests.org/en/latest/api/#requests.delete)
    """

    hook_type: str = 'http_delete'

    # fmt: off
    url: str = Field(..., description="URL for the new request object.")
    kwargs: Union[str, dict] = Field(
        {}, description="Optional arguments that request takes.",
        render_by_default=True
    )
    no_exit: bool = Field(False, description="Whether to exit on non-200 response.")
    # fmt: on

    args: list = ['url', 'kwargs']

    def exec(self):
        r = requests.delete(self.url, auth=self.auth(), **self.kwargs)
        exit_none_200(r, self.no_exit, self.url)

        return r.status_code



import re
from typing import List

from tackle import BaseHook, Field


class UpdateSectionHook(BaseHook):
    """Hook for updating a section of a document."""

    hook_type: str = 'update_section'
    document: str = Field(..., description="Path to a document to render a section of.")
    content: str = Field(
        ..., description="A string to update within the section of the document"
    )
    start_render: str = Field(
        '--start--',
        description="Marker generally in some kind of comment to begin rendering from.",
    )
    end_render: str = Field(
        '--end--',
        description="Marker generally in some kind of comment to end rendering from.",
    )

    args: list = ['document', 'content']

    def split_document(self, doc: List[str]) -> str:
        upper_section = []
        lower_section = []
        in_section = False
        in_upper = True

        for line in doc:
            if re.search(self.start_render, line):
                upper_section.append(line)
                in_section = True
            if re.search(self.end_render, line):
                lower_section.append(line)
                in_section = False
                in_upper = False
            elif not in_section:
                if in_upper:
                    upper_section.append(line)
                else:
                    lower_section.append(line)

        upper_str = "".join(upper_section)
        lower_str = "".join(lower_section)

        return "\n".join([upper_str, self.content, lower_str])

    def exec(self):
        with open(self.document) as f:
            doc = f.readlines()

        output = self.split_document(doc)

        with open(self.document, 'w') as f:
            f.write(output)



"""Test the input source part of the parser."""
import pytest

from tackle.exceptions import (
    EmptyTackleFileException,
    UnknownArgumentException,
    UnknownSourceException,
    HookParseException,
)

# from tackle import tackle
from tackle.cli import main

INPUT_SOURCES = [
    # TODO: empty should now be running help screen
    # ("empty-with-functions.yaml", EmptyTackleFileException),
    ("empty.yaml", EmptyTackleFileException),
    ("out-of-range-arg.yaml", UnknownArgumentException),
    ("non-existent.yaml", UnknownSourceException),
    ("hook-input-validation-error.yaml", HookParseException),
    ("function-input-validation-error.yaml", HookParseException),
]


@pytest.mark.parametrize("input_file,exception", INPUT_SOURCES)
def test_parser_raises_exceptions(chdir, input_file, exception):
    """Test raising exceptions."""
    chdir('exceptions')
    with pytest.raises(exception):
        main([input_file])



import pytest

from tackle import tackle

NON_EXEC_FIXTURES = [
    ('field-hooks.yaml', 'foo'),
    ('field-hooks-method.yaml', 'foo'),
    ('field-hooks-args.yaml', 'bar'),
    ('field-hooks-args-method.yaml', 'bar'),
]


@pytest.mark.parametrize("fixture,expected_output", NON_EXEC_FIXTURES)
def test_function_field_default_with_hooks(chdir, fixture, expected_output):
    """Check that when a declarative hook's default is a hook that it is parsed."""
    chdir('field-hooks-fixtures')
    output = tackle(fixture)

    assert output['call']['literal_compact'] == expected_output
    assert output['call']['literal_expanded'] == expected_output
    assert output['call']['field_default_compact'] == expected_output


EXEC_FIXTURES = [
    ('field-hooks-exec.yaml', 'foo'),
    ('field-hooks-exec-method.yaml', 'foo'),
    ('field-hooks-exec-args.yaml', 'bar'),
    ('field-hooks-exec-args-method.yaml', 'bar'),
]


@pytest.mark.parametrize("fixture,expected_output", EXEC_FIXTURES)
def test_function_field_default_with_hooks_exec(chdir, fixture, expected_output):
    """Check that when a declarative hook's default is a hook that it is parsed."""
    chdir('field-hooks-fixtures')
    output = tackle(fixture)

    assert output['call']['literal_compact_exec'] == expected_output
    assert output['call']['literal_expanded_exec'] == expected_output
    assert output['call']['field_default_compact_exec'] == expected_output


def test_function_field_default_with_hooks_extends(chdir):
    """
    Check that extending a hook works when using a hooks directory with hook field
     default.
    """
    chdir('field-hooks-fixtures')
    output = tackle('extends.yaml')

    assert output['call']['literal_compact'] == 'foo'
    assert output['call']['literal_expanded'] == 'foo'
    assert output['call']['field_default_compact'] == 'foo'


def test_function_field_default_passed_context(chdir):
    """
    Check that when we have hooks with hooks in the default field that the context can
     be passed between the hooks for rendering.
    """
    chdir('field-hooks-fixtures')
    output = tackle('passed-context.yaml')

    assert output['f']['foo'] == 'things'



from tackle import tackle
from tackle.cli import main


# def test_lazy_error(change_curdir_fixtures):
#     main(["lazy.yaml", "-p"])
#
#
# def test_lazy_error_main(change_curdir_fixtures):
#     output = tackle("lazy.yaml")
#     assert output
#
#
# def test_if_issue(change_curdir_fixtures):
#     main(["ifdefined.yaml"])

    # output = tackle("ifdefined.yaml")
    # assert output



"""Tests for function unzip() from zipfile module."""
import tempfile

import pytest

from tackle.utils import zipfile
from tackle.exceptions import InvalidZipRepository


def mock_download():
    """Fake download function."""
    with open('fake-repo-tmpl.zip', 'rb') as zf:
        chunk = zf.read(1024)
        while chunk:
            yield chunk
            chunk = zf.read(1024)


def test_unzip_local_file(mocker, tmpdir, change_curdir_fixtures):
    """Local file reference can be unzipped."""
    mock_prompt_and_delete = mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    output_dir = zipfile.unzip(
        'fake-repo-tmpl.zip',
        clone_to_dir=str(clone_to_dir),
    )

    assert output_dir.startswith(tempfile.gettempdir())
    assert not mock_prompt_and_delete.called


def test_unzip_protected_local_file_environment_password(
    mocker, tmpdir, change_curdir_fixtures
):
    """In `unzip()`, the environment can be used to provide a repo password."""
    mock_prompt_and_delete = mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    output_dir = zipfile.unzip(
        'protected-fake-repo-tmpl.zip',
        clone_to_dir=str(clone_to_dir),
        password='sekrit',
    )

    assert output_dir.startswith(tempfile.gettempdir())
    assert not mock_prompt_and_delete.called


def test_unzip_protected_local_file_bad_environment_password(
    mocker, tmpdir, change_curdir_fixtures
):
    """In `unzip()`, an error occurs if the environment has a bad password."""
    mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    with pytest.raises(InvalidZipRepository):
        zipfile.unzip(
            'protected-fake-repo-tmpl.zip',
            clone_to_dir=str(clone_to_dir),
            password='not-the-right-password',
        )


def test_unzip_protected_local_file_user_password_with_noinput(
    mocker, tmpdir, change_curdir_fixtures
):
    """Can't unpack a password-protected repo in no_input mode."""
    mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    with pytest.raises(InvalidZipRepository):
        zipfile.unzip(
            'protected-fake-repo-tmpl.zip',
            clone_to_dir=str(clone_to_dir),
            no_input=True,
        )


def test_unzip_protected_local_file_user_password(
    mocker, tmpdir, change_curdir_fixtures
):
    """A password-protected local file reference can be unzipped."""
    mock_prompt_and_delete = mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )
    mocker.patch('tackle.utils.zipfile.read_repo_password', return_value='sekrit')

    clone_to_dir = tmpdir.mkdir('clone')

    output_dir = zipfile.unzip(
        'protected-fake-repo-tmpl.zip',
        clone_to_dir=str(clone_to_dir),
    )

    assert output_dir.startswith(tempfile.gettempdir())
    assert not mock_prompt_and_delete.called


def test_unzip_protected_local_file_user_bad_password(
    mocker, tmpdir, change_curdir_fixtures
):
    """Error in `unzip()`, if user can't provide a valid password."""
    mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )
    mocker.patch(
        'tackle.utils.zipfile.read_repo_password',
        return_value='not-the-right-password',
    )

    clone_to_dir = tmpdir.mkdir('clone')

    with pytest.raises(InvalidZipRepository):
        zipfile.unzip(
            'protected-fake-repo-tmpl.zip',
            clone_to_dir=str(clone_to_dir),
        )


def test_empty_zip_file(mocker, tmpdir, change_curdir_fixtures):
    """In `unzip()`, an empty file raises an error."""
    mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    with pytest.raises(InvalidZipRepository):
        zipfile.unzip(
            'empty.zip',
            clone_to_dir=str(clone_to_dir),
        )


def test_non_repo_zip_file(mocker, tmpdir, change_curdir_fixtures):
    """In `unzip()`, a repository must have a top level directory."""
    mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    with pytest.raises(InvalidZipRepository):
        zipfile.unzip(
            'not-a-repo.zip',
            clone_to_dir=str(clone_to_dir),
        )


def test_bad_zip_file(mocker, tmpdir, change_curdir_fixtures):
    """In `unzip()`, a corrupted zip file raises an error."""
    mocker.patch(
        'tackle.utils.zipfile.prompt_and_delete',
        return_value=True,
        autospec=True,
    )

    clone_to_dir = tmpdir.mkdir('clone')

    with pytest.raises(InvalidZipRepository):
        zipfile.unzip(
            'bad-zip-file.zip',
            clone_to_dir=str(clone_to_dir),
        )


# TODO: These tests worked until removing requests dependency.
# Requests could be added back later in which case these can be enabled.

# def test_unzip_url(mocker, tmpdir, change_curdir_fixtures):
#     """In `unzip()`, a url will be downloaded and unzipped."""
#     mock_prompt_and_delete = mocker.patch(
#         'tackle.utils.zipfile.prompt_and_delete',
#         return_value=True,
#         autospec=True,
#     )
#
#     request = mocker.MagicMock()
#     request.iter_content.return_value = mock_download()
#
#     mocker.patch(
#         'tackle.utils.zipfile.requests.get',
#         return_value=request,
#         autospec=True,
#     )
#
#     clone_to_dir = tmpdir.mkdir('clone')
#
#     output_dir = zipfile.unzip(
#         'https://example.com/path/to/fake-repo-tmpl.zip',
#         clone_to_dir=str(clone_to_dir),
#     )
#
#     assert output_dir.startswith(tempfile.gettempdir())
#     assert not mock_prompt_and_delete.called


# def test_unzip_url_existing_cache(mocker, tmpdir, change_curdir_fixtures):
#     """Url should be downloaded and unzipped, old zip file will be removed."""
#     mock_prompt_and_delete = mocker.patch(
#         'tackle.utils.zipfile.prompt_and_delete',
#         return_value=True,
#         autospec=True,
#     )
#
#     request = mocker.MagicMock()
#     request.iter_content.return_value = mock_download()
#
#     mocker.patch(
#         'tackle.utils.zipfile.requests.get',
#         return_value=request,
#         autospec=True,
#     )
#
#     clone_to_dir = tmpdir.mkdir('clone')
#
#     # Create an existing cache of the zipfile
#     existing_zip = clone_to_dir.join('fake-repo-tmpl.zip')
#     existing_zip.write('This is an existing zipfile')
#
#     output_dir = zipfile.unzip(
#         'https://example.com/path/to/fake-repo-tmpl.zip',
#         clone_to_dir=str(clone_to_dir),
#     )
#
#     assert output_dir.startswith(tempfile.gettempdir())
#     assert mock_prompt_and_delete.call_count == 1


# def test_unzip_url_existing_cache_no_input(mocker, tmpdir, change_curdir_fixtures):
#     """If no_input is provided, the existing file should be removed."""
#     request = mocker.MagicMock()
#     request.iter_content.return_value = mock_download()
#
#     mocker.patch(
#         'tackle.utils.zipfile.requests.get',
#         return_value=request,
#         autospec=True,
#     )
#
#     clone_to_dir = tmpdir.mkdir('clone')
#
#     # Create an existing cache of the zipfile
#     existing_zip = clone_to_dir.join('fake-repo-tmpl.zip')
#     existing_zip.write('This is an existing zipfile')
#
#     output_dir = zipfile.unzip(
#         'https://example.com/path/to/fake-repo-tmpl.zip',
#         clone_to_dir=str(clone_to_dir),
#         no_input=True,
#     )
#
#     assert output_dir.startswith(tempfile.gettempdir())


# def test_unzip_should_abort_if_no_redownload(mocker, tmpdir, change_curdir_fixtures):
#     """Should exit without cloning anything If no redownload."""
#     mocker.patch(
#         'tackle.utils.zipfile.prompt_and_delete',
#         side_effect=SystemExit,
#         autospec=True,
#     )
#
#     mock_requests_get = mocker.patch(
#         'tackle.utils.zipfile.requests.get',
#         autospec=True,
#     )
#
#     clone_to_dir = tmpdir.mkdir('clone')
#
#     # Create an existing cache of the zipfile
#     existing_zip = clone_to_dir.join('fake-repo-tmpl.zip')
#     existing_zip.write('This is an existing zipfile')
#
#     zipfile_url = 'https://example.com/path/to/fake-repo-tmpl.zip'
#
#     with pytest.raises(SystemExit):
#         zipfile.unzip(zipfile_url, clone_to_dir=str(clone_to_dir))
#
#     assert not mock_requests_get.called



import sys
import os
import pytest

# from tackle.exceptions import HookCallException

from tackle.main import tackle

if os.name == 'nt':
    pytest.skip("Skipping when run from windows.", allow_module_level=True)


# No idea why this fails - same command...
# def test_provider_system_hook_command(change_dir):
#     context = tackle('list-dir.yaml')
#     assert context['cmd'] == context['cmd_arg']


# TODO: https://github.com/robcxyz/tackle-box/issues/14
# TODO: https://github.com/robcxyz/tackle-box/issues/71
@pytest.mark.skipif(
    sys.platform == 'darwin', reason="https://github.com/robcxyz/tackle-box/issues/71"
)
def test_provider_system_hook_command_multi_line(change_dir):
    output = tackle('multi-line-cmd.yaml')
    assert output['multiline'].startswith('stuff and thing')
    assert output['singleline'].startswith('stuff and thing')


# TODO: https://github.com/robcxyz/tackle-box/issues/13
# def test_provider_system_hook_interactive_shell(chdir):
#     o = tackle('interactive.yaml')
#     assert o


# def test_provider_system_hook_shell_exit(change_dir):
#     with pytest.raises(FileNotFoundError):
#         tackle('exit.yaml')
#
#
# def test_provider_system_hook_shell_exit_long(change_dir):
#     with pytest.raises(HookCallException):
#         tackle('exit-long.yaml')


def test_provider_system_hook_shell_exit_ignore(change_dir):
    o = tackle('exit-ignore.yaml')
    assert o


def test_provider_system_hook_command_exit_ignore(change_dir):
    o = tackle('exit-ignore.yaml')
    assert o



"""Tests for `cookiecutter.utils` module."""
import stat
import sys
from pathlib import Path

import pytest

import tackle.utils.files
import tackle.utils.paths

# import tackle.utils.reader
import tackle


def make_readonly(path):
    """Change the access permissions to readonly for a given file."""
    mode = Path.stat(path).st_mode
    Path.chmod(path, mode & ~stat.S_IWRITE)


@pytest.mark.skipif(
    sys.version_info[0] == 3 and sys.version_info[1] == 6 and sys.version_info[2] == 1,
    reason="Outdated pypy3 version on Travis CI/CD",
)
def test_rmtree(tmp_path):
    """Verify `utils.paths.rmtree` remove files marked as read-only."""
    with open(Path(tmp_path, 'bar'), "w") as f:
        f.write("Test data")
    make_readonly(Path(tmp_path, 'bar'))

    tackle.utils.paths.rmtree(tmp_path)

    assert not Path(tmp_path).exists()


@pytest.mark.skipif(
    sys.version_info[0] == 3 and sys.version_info[1] == 6 and sys.version_info[2] == 1,
    reason="Outdated pypy3 version on Travis CI/CD",
)
def test_make_sure_path_exists(tmp_path):
    """Verify correct True/False response from `utils.paths.make_sure_path_exists`.

    Should return True if directory exist or created.
    Should return False if impossible to create directory (for example protected)
    """
    existing_directory = tmp_path
    directory_to_create = Path(tmp_path, "not_yet_created")

    assert tackle.utils.paths.make_sure_path_exists(existing_directory)
    assert tackle.utils.paths.make_sure_path_exists(directory_to_create)

    # Ensure by base system methods.
    assert existing_directory.is_dir()
    assert existing_directory.exists()
    assert directory_to_create.is_dir()
    assert directory_to_create.exists()


def test_make_sure_path_exists_correctly_handle_os_error(mocker):
    """Verify correct True/False response from `utils.paths.make_sure_path_exists`.

    Should return True if directory exist or created.
    Should return False if impossible to create directory (for example protected)
    """

    def raiser(*args, **kwargs):
        raise OSError()

    mocker.patch("os.makedirs", raiser)
    uncreatable_directory = Path('protected_path')

    assert not tackle.utils.paths.make_sure_path_exists(uncreatable_directory)


@pytest.mark.skipif(
    sys.version_info[0] == 3 and sys.version_info[1] == 6 and sys.version_info[2] == 1,
    reason="Outdated pypy3 version on Travis CI/CD",
)
def test_work_in(tmp_path):
    """Verify returning to original folder after `utils.paths.work_in` use."""
    cwd = Path.cwd()
    ch_to = tmp_path

    assert ch_to != Path.cwd()

    # Under context manager we should work in tmp_path.
    with tackle.utils.paths.work_in(ch_to):
        assert ch_to == Path.cwd()

    # Make sure we return to the correct folder
    assert cwd == Path.cwd()


def test_prompt_should_ask_and_rm_repo_dir(mocker, tmp_path):
    """In `prompt_and_delete()`, if the user agrees to delete/reclone the \
    repo, the repo should be deleted."""
    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no', return_value=True
    )
    repo_dir = Path(tmp_path, 'repo')
    repo_dir.mkdir()

    deleted = tackle.utils.prompts.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert not repo_dir.exists()
    assert deleted


def test_prompt_should_ask_and_exit_on_user_no_answer(mocker, tmp_path):
    """In `prompt_and_delete()`, if the user decline to delete/reclone the \
    repo, cookiecutter should exit."""
    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no',
        return_value=False,
    )
    mock_sys_exit = mocker.patch('sys.exit', return_value=True)
    repo_dir = Path(tmp_path, 'repo')
    repo_dir.mkdir()

    deleted = tackle.utils.prompts.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert repo_dir.exists()
    assert not deleted
    assert mock_sys_exit.called


def test_prompt_should_ask_and_rm_repo_file(mocker, tmp_path):
    """In `prompt_and_delete()`, if the user agrees to delete/reclone a \
    repo file, the repo should be deleted."""
    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no',
        return_value=True,
        autospec=True,
    )

    repo_file = tmp_path.joinpath('repo.zip')
    repo_file.write_text('this is zipfile content')

    deleted = tackle.utils.prompts.prompt_and_delete(str(repo_file))

    assert mock_read_user.called
    assert not repo_file.exists()
    assert deleted


def test_prompt_should_ask_and_keep_repo_on_no_reuse(mocker, tmp_path):
    """In `prompt_and_delete()`, if the user wants to keep their old \
    cloned template repo, it should not be deleted."""
    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no',
        return_value=False,
        autospec=True,
    )
    repo_dir = Path(tmp_path, 'repo')
    repo_dir.mkdir()

    with pytest.raises(SystemExit):
        tackle.utils.prompts.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert repo_dir.exists()


def test_prompt_should_ask_and_keep_repo_on_reuse(mocker, tmp_path):
    """In `prompt_and_delete()`, if the user wants to keep their old \
    cloned template repo, it should not be deleted."""

    def answer(question, default):
        if 'okay to delete' in question:
            return False
        else:
            return True

    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no',
        side_effect=answer,
        autospec=True,
    )
    repo_dir = Path(tmp_path, 'repo')
    repo_dir.mkdir()

    deleted = tackle.utils.prompts.prompt_and_delete(str(repo_dir))

    assert mock_read_user.called
    assert repo_dir.exists()
    assert not deleted


def test_prompt_should_not_ask_if_no_input_and_rm_repo_dir(mocker, tmp_path):
    """Prompt should not ask if no input and rm dir.

    In `prompt_and_delete()`, if `no_input` is True, the call to
    `prompt.read_user_yes_no()` should be suppressed.
    """
    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no',
        return_value=True,
        autospec=True,
    )
    repo_dir = Path(tmp_path, 'repo')
    repo_dir.mkdir()

    deleted = tackle.utils.prompts.prompt_and_delete(str(repo_dir), no_input=True)

    assert not mock_read_user.called
    assert not repo_dir.exists()
    assert deleted


def test_prompt_should_not_ask_if_no_input_and_rm_repo_file(mocker, tmp_path):
    """Prompt should not ask if no input and rm file.

    In `prompt_and_delete()`, if `no_input` is True, the call to
    `prompt.read_user_yes_no()` should be suppressed.
    """
    mock_read_user = mocker.patch(
        'tackle.utils.prompts.read_user_yes_no',
        return_value=True,
        autospec=True,
    )

    repo_file = tmp_path.joinpath('repo.zip')
    repo_file.write_text('this is zipfile content')

    deleted = tackle.utils.prompts.prompt_and_delete(str(repo_file), no_input=True)

    assert not mock_read_user.called
    assert not repo_file.exists()
    assert deleted


@pytest.mark.parametrize(
    'valid_config_file',
    (
        [
            'valid/tackle-input/tackle.yml',
            'valid/yaml-input/cookiecutter.yaml',
        ]
    ),
)
def test_valid_read_config_file(valid_config_file, change_curdir_fixtures):
    """Validate generic reader works properly."""
    output = tackle.utils.files.read_config_file(valid_config_file)
    assert output == {'project_slug': 'best_eva', 'stuff': 'things'}



# from typing import Union
#
# from tackle.models import BaseHook, Field
#
#
# class MatchHook(BaseHook):
#     """
#     Hook for while loops. Takes a condition and runs a context in a loop until the
#     condition is met.
#
#     Not Implemented yet.
#     """
#
#     hook_type: str = 'while'
#     condition: str = Field(
#         ..., description="A jinja expression to evaluated on each pass through `run`."
#     )
#     run: Union[dict, list] = Field(
#         ...,
#         description="A list or dictionary to parse same as tackle file.",
#     )
#
#     args: list = ['condition']
#
#     def exec(self) -> Union[dict, list]:
#         raise NotImplementedError



# from tackle.models import BaseHook, Field
# from requests import get
#
# class GitHubRawHook(BaseHook):
#     """
#     Hook for github raw content.
#     """
#
#     hook_type: str = 'github_raw'
#
#     org: str = Field(None, description="The github username or org.")
#     repo: str = Field(None, description="The repo name.")
#     file_path: str = Field(None, description="")
#
#     output_path: str = Field(None, description="Optional param to take output and write to file location.")
#
#     args: list = ['org', 'repo']
#
#     def make_url(self) -> str:
#         url = f"https://raw.githubusercontent.com/{self.org}/{self.repo}/master/path/to/file "
#         return url
#
#     def make_output_path(self) -> str:
#         if self.output_path == ".":
#             self.
#
#
#     def exec(self):
#         url = self.make_url()
#         output = get(url=url)
#
#         if self.output_path is not None:
#             with open(self.output_path, 'w') as f:
#                 f.write()
#



import sys
from InquirerPy import prompt

from typing import Any
from tackle.models import BaseHook, Field
from tackle.utils.dicts import get_readable_key_path


class InquirerRawListHook(BaseHook):
    """
    Hook for PyInquirer 'rawlist' type prompts. Similar to `select` hook with less
    flexibility. [Source example](https://github.com/CITGuru/PyInquirer/blob/master/examples/rawlist.py)
    """

    hook_type: str = 'rawlist'

    default: Any = Field(None, description="Default choice.")
    message: str = Field(None, description="String message to show when prompting.")
    choices: list = Field(..., description="A list of choices.")

    args: list = ['message', 'default']

    def exec(self) -> list:
        if self.message is None:
            self.message = get_readable_key_path(self.key_path) + ' >>>'

        if not self.no_input:
            question = {
                'type': self.hook_type,
                'name': 'tmp',
                'message': self.message,
                'choices': self.choices,
            }
            if self.default:
                question.update({'default': self.default})

            # Handle keyboard exit
            try:
                response = prompt([question])
            except KeyboardInterrupt:
                print("Exiting...")
                sys.exit(0)
            return response['tmp']

        elif self.default:
            return self.default
        else:
            # When no_input then return empty list
            return []



# from tackle import BaseHook, Field
#
#
# class MapListKeyValuesHook(BaseHook):
#     """
#     Hook for creating a list of maps from a map with keys as `key` + `value` from the
#      map. Useful for some operations that require maps as lists with these keys.
#     """
#
#     hook_type: str = 'map_to_list_key_values'
#     # fmt: off
#     src: dict = Field(
#         ..., description="A map to extract the keys out of.", render_by_default=True)
#     # fmt: on
#
#     args: list = ['src']
#
#     def exec(self) -> list:
#         output = []
#         for k, v in self.src.items():
#             output.append({'key': k, 'value': v})
#
#         return output



from tackle.main import tackle


def test_provider_console_print(change_dir):
    output = tackle('print.yaml')
    assert output


def test_provider_console_pprint(change_dir):
    output = tackle('pprint.yaml')
    assert 'this' in output



import pytest

from tackle import tackle

# from tackle.parser import handle_leading_brackets
from tackle.macros import var_hook_macro
from tackle.utils.command import unpack_args_kwargs_string

TEMPLATES = [
    # These should all have `var` prepended to the args as it is indicative of a render
    ('"{{foo}}" bar baz', 4, 0, 0),
    ('{{ foo }} bar baz', 4, 0, 0),
    ('{{ foo }} bar baz bing', 5, 0, 0),
    ('{{ foo}} bar baz', 4, 0, 0),
    ('{{foo }} bar baz', 4, 0, 0),
    ('{{ foo }}-foo bar baz', 4, 0, 0),
    ('bar-{{ foo }}-foo', 2, 0, 0),
    ('bar-{{ foo in var }}-foo', 2, 0, 0),
]


@pytest.mark.parametrize("template,len_args,len_kwargs,len_flags", TEMPLATES)
def test_unpack_args_kwargs_handle_leading_brackets(
    template, len_args, len_kwargs, len_flags
):
    """Validate the count of each input arg/kwarg/flag."""
    args, kwargs, flags = unpack_args_kwargs_string(template)
    args = var_hook_macro(args)

    assert len_args == len(args)
    assert len_kwargs == len(kwargs)
    assert len_flags == len(flags)


def test_parser_types(change_curdir_fixtures):
    """Check that types are respected."""
    output = tackle('types.yaml')
    assert output['hex'] == '0x01'
    assert output['float'] == 1.0
    assert output['bool']

    assert isinstance(output['int'], int)
    assert isinstance(output['float'], float)
    assert isinstance(output['hex'], str)
    assert isinstance(output['bool'], bool)

    assert output['bool'] == output['bool_render']
    assert output['int'] == output['int_render']
    assert output['float'] == output['float_render']
    assert output['hex'] == output['hex_render']



"""Test the cli."""
import pytest
import os

from tackle.models import Context
from tackle.cli import main

# fmt: off
INPUT_SOURCES = [
    ("github.com/robcxyz/tackle-demo",
     {'input_string': "github.com/robcxyz/tackle-demo"}),
    ("robcxyz/tackle-demo --no-input",
     {'input_string': 'robcxyz/tackle-demo', 'no_input': True}),
    ("thing --no-input", {'input_string': 'thing', 'no_input': True}),
    ("thing --foo bar", {'input_string': 'thing', 'global_kwargs': {'foo': 'bar'}}),
    ("thing foo bar", {'input_string': 'thing', 'global_args': ['foo', 'bar']}),
]
# fmt: on


@pytest.mark.parametrize("input_string,output", INPUT_SOURCES)
def test_cli_parse_args(mocker, change_dir_base, input_string, output):
    """Mock the main call and verify the args get passed in right through the CLI."""
    mock = mocker.patch("tackle.main.update_source", autospec=True, return_value={})
    main(input_string.split(' '))

    assert mock.called
    assert isinstance(mock.call_args[0][0], Context)

    context = mock.call_args[0][0].dict()

    for k, v in output.items():
        assert k in context
        assert context[k] == v


def test_cli_parse_args_empty(mocker, change_curdir_fixtures):
    """When no arg is given we should find the closest tackle file."""
    mock = mocker.patch("tackle.main.update_source", autospec=True, return_value={})
    main([])

    assert mock.called
    assert isinstance(mock.call_args[0][0], Context)
    context = mock.call_args[0][0].dict()
    assert context['input_string'] == os.path.abspath('.tackle.yaml')


PRINTS = ["--print", "-p"]


@pytest.mark.parametrize("input_string", PRINTS)
def test_cli_parse_args_print_option(
    mocker, change_curdir_fixtures, capsys, input_string
):
    """When no arg is given we should find the closest tackle file."""
    mocker.patch("tackle.main.tackle", autospec=True, return_value={})
    main([input_string])
    assert '{"stuff": "things"}' in capsys.readouterr().out


def test_cli_parse_args_help():
    """Test help arg."""
    with pytest.raises(SystemExit) as e:
        main(["--help"])
        assert e.value.code == 0

    with pytest.raises(SystemExit) as e:
        main(["-h"])
        assert e.value.code == 0


def test_cli_parse_args_version():
    """Test version arg."""
    with pytest.raises(SystemExit) as e:
        main(["--version"])
        assert e.value.code == 0



import json
import os
from typing import Union

from tackle.models import BaseHook, Field


class JsonHook(BaseHook):
    """
    Hook for reading and writing json. Hook reads from `path` if no `data` field is
     provided, otherwise it writes the `data` to `path`.
    """

    hook_type: str = 'json'
    path: str = Field(..., description="The file path to put read or write to.")
    data: Union[dict, list, str] = Field(
        None,
        description="Map/list or renderable string to a map/list key to write.",
        render_by_default=True,
    )
    args: list = ['path', 'data']

    def exec(self) -> Union[dict, str]:
        self.path = os.path.abspath(os.path.expanduser(os.path.expandvars(self.path)))
        # Make path if it does not exist
        if not os.path.exists(os.path.dirname(self.path)) and self.data:
            os.makedirs(os.path.dirname(self.path))

        if self.data:
            with open(self.path, 'w') as f:
                json.dump(self.data, f)
            return self.path
        else:
            with open(self.path, 'r') as f:
                data = json.load(f)
            return data


class JsonifyHook(BaseHook):
    """
    Hook for reading and writing json. Hook reads from `path` if no `data` field is
     provided, otherwise it writes the `data` to `path`.
    """

    hook_type: str = 'jsonify'
    data: Union[dict, list, str] = Field(
        ...,
        description="Map/list or renderable string to a map/list key to write.",
        render_by_default=True,
    )
    args: list = ['data']

    def exec(self) -> Union[dict, str]:
        return json.dumps(self.data)
        # if self.path:
        #     self.path = os.path.abspath(
        #         os.path.expanduser(os.path.expandvars(self.path))
        #     )
        #     # Make path if it does not exist
        #     if not os.path.exists(os.path.dirname(self.path)) and self.data:
        #         os.makedirs(os.path.dirname(self.path))
        #     with open(self.path, 'w') as f:
        #         json.dump(self.data, f)
        #     return self.path
        # else:
        #     return json.dumps(self.data)
